# Copyright (c) 2023 Darren Erik Vengroff
"""Places in Kansas."""


ABBYVILLE_CITY = "00100"
ABILENE_CITY = "00125"
ADA_CDP = "00225"
ADMIRE_CITY = "00325"
AGENDA_CITY = "00475"
AGRA_CITY = "00550"
ALBERT_CITY = "00825"
ALDEN_CITY = "00950"
ALEXANDER_CITY = "01050"
ALLEN_CITY = "01275"
ALMA_CITY = "01350"
ALMENA_CITY = "01425"
ALTA_VISTA_CITY = "01575"
ALTAMONT_CITY = "01550"
ALTON_CITY = "01600"
ALTOONA_CITY = "01625"
AMERICUS_CITY = "01675"
AMES_CDP = "01725"
ANDALE_CITY = "01775"
ANDOVER_CITY = "01800"
ANTHONY_CITY = "01975"
ARCADIA_CITY = "02175"
ARGONIA_CITY = "02250"
ARKANSAS_CITY_CITY = "02300"
ARLINGTON_CITY = "02325"
ARMA_CITY = "02375"
ASHERVILLE_CDP = "02600"
ASHLAND_CITY = "02675"
ASSARIA_CITY = "02850"
ATCHISON_CITY = "02900"
ATHOL_CITY = "03025"
ATLANTA_CITY = "03050"
ATTICA_CITY = "03100"
ATWOOD_CITY = "03150"
AUBURN_CITY = "03250"
AUGUSTA_CITY = "03300"
AURORA_CITY = "03425"
AXTELL_CITY = "03600"
BAILEYVILLE_CDP = "03675"
BALA_CDP = "03800"
BALDWIN_CITY_CITY = "03900"
BARNARD_CITY = "04225"
BARNES_CITY = "04250"
BARTLETT_CITY = "04350"
BASEHOR_CITY = "04400"
BASSETT_CITY = "04450"
BAVARIA_CDP = "04575"
BAXTER_SPRINGS_CITY = "04625"
BAZINE_CITY = "04775"
BEATTIE_CITY = "04900"
BEAUMONT_CDP = "04925"
BEAVER_CDP = "04950"
BEL_AIRE_CITY = "05337"
BELLE_PLAINE_CITY = "05500"
BELLEVILLE_CITY = "05600"
BELOIT_CITY = "05775"
BELPRE_CITY = "05825"
BELVUE_CITY = "05925"
BENDENA_CDP = "05975"
BENEDICT_CITY = "06000"
BENNINGTON_CITY = "06075"
BENTLEY_CITY = "06125"
BENTON_CITY = "06175"
BERN_CITY = "06275"
BEVERLY_CITY = "06500"
BIG_BOW_CDP = "06575"
BIRD_CITY_CITY = "06825"
BISON_CITY = "06950"
BLUE_MOUND_CITY = "07600"
BLUE_RAPIDS_CITY = "07650"
BLUFF_CITY_CITY = "07775"
BOGUE_CITY = "07825"
BONNER_SPRINGS_CITY = "07975"
BREMEN_CDP = "08300"
BREWSTER_CITY = "08425"
BRIDGEPORT_CDP = "08450"
BRONSON_CITY = "08475"
BROOKVILLE_CITY = "08575"
BROWNELL_CITY = "08675"
BUCKLIN_CITY = "09000"
BUCYRUS_CDP = "09050"
BUFFALO_CITY = "09150"
BUHLER_CITY = "09200"
BUNKER_HILL_CITY = "09225"
BURDEN_CITY = "09250"
BURDETT_CITY = "09275"
BURDICK_CDP = "09300"
BURLINGAME_CITY = "09350"
BURLINGTON_CITY = "09400"
BURNS_CITY = "09450"
BURR_OAK_CITY = "09525"
BURRTON_CITY = "09575"
BUSHONG_CITY = "09675"
BUSHTON_CITY = "09700"
BYERS_CITY = "09800"
CALDWELL_CITY = "09900"
CAMBRIDGE_CITY = "10125"
CANEY_CITY = "10375"
CANTON_CITY = "10475"
CARBONDALE_CITY = "10600"
CARLTON_CITY = "10675"
CASSODAY_CITY = "10925"
CATHARINE_CDP = "11025"
CAWKER_CITY_CITY = "11175"
CEDAR_POINT_CITY = "11425"
CEDAR_VALE_CITY = "11450"
CEDAR_CITY = "11325"
CENTERVILLE_CDP = "12325"
CENTRALIA_CITY = "12425"
CENTROPOLIS_CDP = "12450"
CHANUTE_CITY = "12500"
CHAPMAN_CITY = "12550"
CHASE_CITY = "12650"
CHAUTAUQUA_CITY = "12675"
CHENEY_CITY = "12775"
CHEROKEE_CITY = "12825"
CHERRYVALE_CITY = "12925"
CHETOPA_CITY = "12950"
CHICOPEE_CDP = "13100"
CIMARRON_CITY = "13275"
CIRCLEVILLE_CITY = "13375"
CLAFLIN_CITY = "13425"
CLAY_CENTER_CITY = "13625"
CLAYTON_CITY = "13675"
CLEARWATER_CITY = "13925"
CLIFTON_CITY = "14200"
CLIMAX_CITY = "14275"
CLYDE_CITY = "14475"
COATS_CITY = "14525"
CODELL_CDP = "14575"
COFFEYVILLE_CITY = "14600"
COLBY_CITY = "14650"
COLDWATER_CITY = "14700"
COLLYER_CITY = "14900"
COLONY_CITY = "14950"
COLUMBUS_CITY = "15075"
COLWICH_CITY = "15100"
CONCORDIA_CITY = "15200"
CONWAY_SPRINGS_CITY = "15325"
COOLIDGE_CITY = "15400"
COPELAND_CITY = "15475"
CORNING_CITY = "15725"
COTTONWOOD_FALLS_CITY = "15900"
COUNCIL_GROVE_CITY = "15925"
COURTLAND_CITY = "16025"
COYVILLE_CITY = "16150"
CRESTLINE_CDP = "16325"
CROWEBURG_CDP = "16475"
CUBA_CITY = "16625"
CULLISON_CITY = "16675"
CULVER_CITY = "16700"
CUNNINGHAM_CITY = "16775"
DAMAR_CITY = "16975"
DANVILLE_CITY = "17000"
DE_SOTO_CITY = "17850"
DEARING_CITY = "17150"
DEERFIELD_CITY = "17225"
DELIA_CITY = "17525"
DELPHOS_CITY = "17600"
DENISON_CITY = "17625"
DENNIS_CDP = "17675"
DENTON_CITY = "17750"
DERBY_CITY = "17800"
DETROIT_CDP = "17875"
DEVON_CDP = "17900"
DEXTER_CITY = "17925"
DIGHTON_CITY = "18050"
DODGE_CITY_CITY = "18250"
DORRANCE_CITY = "18325"
DOUGLASS_CITY = "18400"
DOWNS_CITY = "18500"
DRESDEN_CITY = "18600"
DUNLAP_CITY = "18950"
DURHAM_CITY = "19050"
DWIGHT_CITY = "19125"
EARLTON_CITY = "19275"
EASTBOROUGH_CITY = "19300"
EASTON_CITY = "19600"
EASTSHORE_CDP = "19700"
EDGERTON_CITY = "19825"
EDMOND_CITY = "19875"
EDNA_CITY = "19900"
EDSON_CDP = "19925"
EDWARDSVILLE_CITY = "20000"
EFFINGHAM_CITY = "20025"
EL_DORADO_CITY = "20075"
ELBING_CITY = "20050"
ELGIN_CITY = "20125"
ELK_CITY_CITY = "20250"
ELK_FALLS_CITY = "20300"
ELKHART_CITY = "20350"
ELLINWOOD_CITY = "20425"
ELLIS_CITY = "20450"
ELLSWORTH_CITY = "20500"
ELMDALE_CITY = "20675"
ELSMORE_CITY = "20875"
ELWOOD_CITY = "20950"
ELYRIA_CDP = "20975"
EMMETT_CITY = "21100"
EMPORIA_CITY = "21275"
ENGLEWOOD_CITY = "21350"
ENSIGN_CITY = "21400"
ENTERPRISE_CITY = "21425"
ERIE_CITY = "21500"
ESBON_CITY = "21600"
ESKRIDGE_CITY = "21650"
EUDORA_CITY = "21675"
EUREKA_CITY = "21800"
EVEREST_CITY = "22025"
FAIRVIEW_CITY = "22425"
FAIRWAY_CITY = "22700"
FALL_RIVER_CITY = "22775"
FALUN_CDP = "22900"
FARLINGTON_CDP = "23050"
FLORENCE_CITY = "23600"
FONTANA_CITY = "23650"
FORD_CITY = "23725"
FORMOSO_CITY = "23825"
FORT_DODGE_CDP = "23890"
FORT_RILEY_CDP = "23975"
FORT_SCOTT_CITY = "24000"
FOWLER_CITY = "24175"
FRANKFORT_CITY = "24275"
FRANKLIN_CDP = "24325"
FREDERICK_CITY = "24550"
FREDONIA_CITY = "24575"
FRONTENAC_CITY = "24850"
FULTON_CITY = "24925"
FURLEY_CDP = "24975"
GALATIA_CITY = "25050"
GALENA_CITY = "25100"
GALESBURG_CITY = "25150"
GALVA_CITY = "25200"
GARDEN_CITY_CITY = "25325"
GARDEN_PLAIN_CITY = "25375"
GARDNER_CITY = "25425"
GARFIELD_CITY = "25750"
GARLAND_CDP = "25900"
GARNETT_CITY = "25925"
GAS_CITY = "25975"
GAYLORD_CITY = "26000"
GEM_CITY = "26050"
GENESEO_CITY = "26075"
GEUDA_SPRINGS_CITY = "26200"
GIRARD_CITY = "26300"
GLADE_CITY = "26325"
GLASCO_CITY = "26375"
GLEN_ELDER_CITY = "26500"
GODDARD_CITY = "26725"
GOESSEL_CITY = "26750"
GOFF_CITY = "26775"
GOODLAND_CITY = "26875"
GORHAM_CITY = "27000"
GOVE_CITY_CITY = "27125"
GRAINFIELD_CITY = "27200"
GRANDVIEW_PLAZA_CITY = "27425"
GRANTVILLE_CDP = "28200"
GREAT_BEND_CITY = "28300"
GREELEY_COUNTY_UNIFIED_GOVERNMENT_BALANCE = "28412"
GREELEY_CITY = "28350"
GREEN_CITY = "28425"
GREENLEAF_CITY = "28625"
GREENSBURG_CITY = "28675"
GREENWICH_CDP = "28700"
GRENOLA_CITY = "28800"
GRIDLEY_CITY = "28850"
GRINNELL_CITY = "28900"
GYPSUM_CITY = "29250"
HADDAM_CITY = "29400"
HALLOWELL_CDP = "29550"
HALSTEAD_CITY = "29600"
HAMILTON_CITY = "29675"
HAMLIN_CITY = "29700"
HANOVER_CITY = "29875"
HANSTON_CITY = "29925"
HARDTNER_CITY = "30000"
HARPER_CITY = "30175"
HARRIS_CDP = "30250"
HARTFORD_CITY = "30475"
HARVEYVILLE_CITY = "30575"
HAVANA_CITY = "30700"
HAVEN_CITY = "30725"
HAVENSVILLE_CITY = "30775"
HAVILAND_CITY = "30825"
HAYS_CITY = "31100"
HAYSVILLE_CITY = "31125"
HAZELTON_CITY = "31150"
HEALY_CDP = "31200"
HEPLER_CITY = "31375"
HERINGTON_CITY = "31400"
HERKIMER_CDP = "31425"
HERNDON_CITY = "31475"
HESSTON_CITY = "31600"
HIAWATHA_CITY = "31675"
HIGHLAND_CITY = "31850"
HILL_CITY_CITY = "32175"
HILLSBORO_CITY = "32275"
HILLSDALE_CDP = "32300"
HOISINGTON_CITY = "32550"
HOLCOMB_CITY = "32575"
HOLLENBERG_CITY = "32650"
HOLTON_CITY = "32825"
HOLYROOD_CITY = "32850"
HOME_CDP = "32875"
HOPE_CITY = "33075"
HORACE_CITY = "33150"
HORTON_CITY = "33200"
HOWARD_CITY = "33250"
HOXIE_CITY = "33350"
HOYT_CITY = "33375"
HUDSON_CITY = "33400"
HUGOTON_CITY = "33425"
HUMBOLDT_CITY = "33450"
HUNNEWELL_CITY = "33500"
HUNTER_CITY = "33525"
HURON_CITY = "33575"
HUTCHINSON_CITY = "33625"
IDANA_CDP = "33700"
INDEPENDENCE_CITY = "33875"
INGALLS_CITY = "34225"
INMAN_CITY = "34275"
IOLA_CITY = "34300"
IONIA_CDP = "34350"
ISABEL_CITY = "34550"
IUKA_CITY = "34625"
JAMESTOWN_CITY = "34975"
JENNINGS_CITY = "35375"
JETMORE_CITY = "35450"
JEWELL_CITY = "35475"
JOHNSON_CITY_CITY = "35600"
JUNCTION_CITY_CITY = "35750"
KANOPOLIS_CITY = "35950"
KANORADO_CITY = "35975"
KANSAS_CITY_CITY = "36000"
KEATS_CDP = "36200"
KECHI_CITY = "36225"
KELLY_CDP = "36375"
KENSINGTON_CITY = "36550"
KICKAPOO_SITE_1_CDP = "36735"
KICKAPOO_SITE_2_CDP = "36736"
KICKAPOO_SITE_5_CDP = "36739"
KICKAPOO_SITE_6_CDP = "36740"
KICKAPOO_SITE_7_CDP = "36741"
KICKAPOO_TRIBAL_CENTER_CDP = "36744"
KINCAID_CITY = "36875"
KINGMAN_CITY = "36950"
KINSLEY_CITY = "37075"
KIOWA_CITY = "37125"
KIPP_CDP = "37200"
KIRWIN_CITY = "37275"
KISMET_CITY = "37325"
LA_CROSSE_CITY = "37500"
LA_CYGNE_CITY = "37575"
LA_HARPE_CITY = "37725"
LABETTE_CITY = "37375"
LAFONTAINE_CDP = "37675"
LAKE_CITY_CDP = "37875"
LAKE_QUIVIRA_CITY = "37975"
LAKIN_CITY = "38175"
LANCASTER_CITY = "38325"
LANE_CITY = "38400"
LANGDON_CITY = "38500"
LANSING_CITY = "38650"
LARNED_CITY = "38700"
LATHAM_CITY = "38775"
LATIMER_CITY = "38800"
LAWRENCE_CITY = "38900"
LEROY_CITY = "39650"
LEAVENWORTH_CITY = "39000"
LEAWOOD_CITY = "39075"
LEBANON_CITY = "39100"
LEBO_CITY = "39125"
LECOMPTON_CITY = "39150"
LEHIGH_CITY = "39225"
LENEXA_CITY = "39350"
LENORA_CITY = "39400"
LEON_CITY = "39450"
LEONA_CITY = "39475"
LEONARDVILLE_CITY = "39500"
LEOTI_CITY = "39550"
LEVANT_CDP = "39700"
LEWIS_CITY = "39725"
LIBERAL_CITY = "39825"
LIBERTY_CITY = "40250"
LIEBENTHAL_CITY = "40400"
LINCOLN_CENTER_CITY = "41300"
LINCOLNVILLE_CITY = "41325"
LINDSBORG_CITY = "41375"
LINN_VALLEY_CITY = "41465"
LINN_CITY = "41425"
LINWOOD_CITY = "41475"
LITTLE_RIVER_CITY = "41575"
LOGAN_CITY = "42100"
LONE_ELM_CITY = "42425"
LONG_ISLAND_CITY = "42600"
LONGFORD_CITY = "42575"
LONGTON_CITY = "42650"
LORRAINE_CITY = "42800"
LOST_SPRINGS_CITY = "42825"
LOUISBURG_CITY = "42875"
LOUISVILLE_CITY = "42925"
LOWELL_CDP = "43050"
LUCAS_CITY = "43150"
LUDELL_CDP = "43175"
LURAY_CITY = "43250"
LYNDON_CITY = "43350"
LYONS_CITY = "43525"
MACKSVILLE_CITY = "43850"
MADISON_CITY = "44050"
MAHASKA_CITY = "44150"
MAIZE_CITY = "44200"
MANCHESTER_CITY = "44225"
MANHATTAN_CITY = "44250"
MANKATO_CITY = "44300"
MANTER_CITY = "44375"
MAPLE_HILL_CITY = "44500"
MAPLETON_CITY = "44550"
MARIENTHAL_CDP = "44600"
MARION_CITY = "44750"
MARQUETTE_CITY = "44925"
MARYSVILLE_CITY = "45050"
MATFIELD_GREEN_CITY = "45150"
MAYETTA_CITY = "45250"
MAYFIELD_CITY = "45275"
MCCONNELL_AFB_CDP = "43675"
MCCRACKEN_CITY = "43700"
MCCUNE_CITY = "43725"
MCDONALD_CITY = "43750"
MCFARLAND_CITY = "43775"
MCLOUTH_CITY = "43900"
MCPHERSON_CITY = "43950"
MEADE_CITY = "45325"
MEDICINE_LODGE_CITY = "45500"
MELVERN_CITY = "45700"
MENLO_CITY = "45750"
MENTOR_CDP = "45875"
MERIDEN_CITY = "45950"
MERRIAM_CITY = "46000"
MILAN_CITY = "46450"
MILDRED_CDP = "46500"
MILFORD_CITY = "46525"
MILTON_CDP = "46925"
MILTONVALE_CITY = "46950"
MINNEAPOLIS_CITY = "47075"
MINNEOLA_CITY = "47150"
MISSION_HILLS_CITY = "47350"
MISSION_WOODS_CITY = "47425"
MISSION_CITY = "47225"
MOLINE_CITY = "47600"
MONT_IDA_CDP = "47975"
MONTEZUMA_CITY = "47875"
MONUMENT_CDP = "48025"
MORAN_CITY = "48125"
MORGANVILLE_CITY = "48225"
MORLAND_CITY = "48275"
MORRILL_CITY = "48300"
MORROWVILLE_CITY = "48450"
MOSCOW_CITY = "48600"
MOUND_CITY_CITY = "48750"
MOUND_VALLEY_CITY = "48825"
MOUNDRIDGE_CITY = "48800"
MOUNT_HOPE_CITY = "48900"
MULBERRY_CITY = "49025"
MULLINVILLE_CITY = "49075"
MULVANE_CITY = "49100"
MUNDEN_CITY = "49150"
MUNJOR_CDP = "49200"
MURDOCK_CDP = "49250"
MUSCOTAH_CITY = "49325"
NARKA_CITY = "49350"
NASHVILLE_CITY = "49400"
NATOMA_CITY = "49425"
NAVARRE_CDP = "49525"
NEAL_CDP = "49550"
NEODESHA_CITY = "49650"
NEOSHO_FALLS_CITY = "49825"
NEOSHO_RAPIDS_CITY = "49875"
NESS_CITY_CITY = "49925"
NETAWAKA_CITY = "49950"
NEW_ALBANY_CITY = "50150"
NEW_CAMBRIA_CITY = "50300"
NEW_SALEM_CDP = "50425"
NEW_STRAWN_CITY = "50450"
NEWBURY_CDP = "50250"
NEWTON_CITY = "50475"
NICKERSON_CITY = "50525"
NICODEMUS_CDP = "50550"
NILES_CDP = "50600"
NIOTAZE_CITY = "50750"
NORCATUR_CITY = "50925"
NORTH_NEWTON_CITY = "51225"
NORTON_CITY = "51500"
NORTONVILLE_CITY = "51525"
NORWAY_CDP = "51550"
NORWICH_CITY = "51600"
OAK_HILL_CITY = "51700"
OAKLAWN_SUNVIEW_CDP = "51810"
OAKLEY_CITY = "51825"
OBERLIN_CITY = "52000"
ODIN_CDP = "52175"
OFFERLE_CITY = "52200"
OGALLAH_CDP = "52225"
OGDEN_CITY = "52275"
OKETO_CITY = "52525"
OLATHE_CITY = "52575"
OLIVET_CITY = "52700"
OLMITZ_CITY = "52750"
OLPE_CITY = "52800"
OLSBURG_CITY = "52825"
ONAGA_CITY = "52875"
ONEIDA_CITY = "52900"
OPOLIS_CDP = "52925"
OSAGE_CITY_CITY = "53200"
OSAWATOMIE_CITY = "53225"
OSBORNE_CITY = "53325"
OSKALOOSA_CITY = "53375"
OSWEGO_CITY = "53450"
OTIS_CITY = "53525"
OTTAWA_CITY = "53550"
OVERBROOK_CITY = "53725"
OVERLAND_PARK_CITY = "53775"
OXFORD_CITY = "53850"
OZAWKIE_CITY = "53925"
PALCO_CITY = "54125"
PALMER_CITY = "54200"
PAOLA_CITY = "54250"
PARADISE_CITY = "54325"
PARK_CITY_CITY = "54450"
PARK_CITY = "54400"
PARKER_CITY = "54500"
PARKERFIELD_CITY = "54560"
PARKERVILLE_CITY = "54575"
PARSONS_CITY = "54675"
PARTRIDGE_CITY = "54700"
PAWNEE_ROCK_CITY = "54875"
PAXICO_CITY = "54950"
PEABODY_CITY = "55100"
PECK_CDP = "55175"
PENALOSA_CITY = "55225"
PERRY_CITY = "55450"
PERU_CITY = "55525"
PHILLIPSBURG_CITY = "55675"
PIEDMONT_CDP = "55750"
PIERCEVILLE_CDP = "55775"
PILSEN_CDP = "55875"
PIQUA_CDP = "56000"
PITTSBURG_CITY = "56025"
PLAINS_CITY = "56100"
PLAINVILLE_CITY = "56150"
PLEASANTON_CITY = "56450"
PLEVNA_CITY = "56725"
POMONA_CITY = "57000"
PORTIS_CITY = "57100"
POTWIN_CITY = "57300"
POWHATTAN_CITY = "57350"
PRAIRIE_VIEW_CITY = "57525"
PRAIRIE_VILLAGE_CITY = "57575"
PRATT_CITY = "57625"
PRESCOTT_CITY = "57650"
PRESTON_CITY = "57675"
PRETTY_PRAIRIE_CITY = "57700"
PRINCETON_CITY = "57725"
PROTECTION_CITY = "57850"
QUENEMO_CITY = "58100"
QUINTER_CITY = "58250"
RADIUM_CITY = "58300"
RADLEY_CDP = "58325"
RAMONA_CITY = "58375"
RANDALL_CITY = "58450"
RANDOLPH_CITY = "58475"
RANSOM_CITY = "58500"
RANTOUL_CITY = "58525"
RAYMOND_CITY = "58550"
READING_CITY = "58600"
REDFIELD_CITY = "58725"
REPUBLIC_CITY = "59000"
RESERVE_CITY = "59050"
REXFORD_CITY = "59125"
RICHFIELD_CITY = "59200"
RICHMOND_CITY = "59675"
RILEY_CITY = "59875"
RINGO_CDP = "59975"
RIVERTON_CDP = "60200"
ROBINSON_CITY = "60325"
ROCK_CDP = "60400"
ROELAND_PARK_CITY = "60825"
ROLLA_CITY = "60900"
ROSALIA_CDP = "61100"
ROSE_HILL_CITY = "61250"
ROSELAND_CITY = "61275"
ROSSVILLE_CITY = "61400"
ROXBURY_CDP = "61550"
ROZEL_CITY = "61600"
RUSH_CENTER_CITY = "61775"
RUSSELL_SPRINGS_CITY = "61875"
RUSSELL_CITY = "61825"
SABETHA_CITY = "62025"
SALINA_CITY = "62700"
SATANTA_CITY = "63150"
SAVONBURG_CITY = "63200"
SAWYER_CITY = "63275"
SCAMMON_CITY = "63325"
SCANDIA_CITY = "63350"
SCHOENCHEN_CITY = "63425"
SCOTT_CITY_CITY = "63600"
SCOTTSVILLE_CITY = "63650"
SCRANTON_CITY = "63675"
SEDAN_CITY = "63750"
SEDGWICK_CITY = "63800"
SELDEN_CITY = "63875"
SENECA_CITY = "63950"
SEVERANCE_CITY = "64025"
SEVERY_CITY = "64050"
SEWARD_CITY = "64100"
SHALLOW_WATER_CDP = "64200"
SHARON_SPRINGS_CITY = "64375"
SHARON_CITY = "64325"
SHAWNEE_CITY = "64500"
SILVER_LAKE_CITY = "65600"
SILVERDALE_CDP = "65550"
SIMPSON_CITY = "65650"
SMITH_CENTER_CITY = "65925"
SMOLAN_CITY = "66100"
SOLDIER_CITY = "66175"
SOLOMON_CITY = "66275"
SOMERSET_CDP = "66475"
SOUTH_HAVEN_CITY = "66650"
SOUTH_HUTCHINSON_CITY = "66750"
SOUTH_MOUND_CDP = "66800"
SPEARVILLE_CITY = "67125"
SPEED_CITY = "67175"
SPIVEY_CITY = "67250"
SPRING_HILL_CITY = "67625"
ST_BENEDICT_CDP = "62075"
ST_FRANCIS_CITY = "62175"
ST_GEORGE_CITY = "62200"
ST_JOHN_CITY = "62275"
ST_MARKS_CDP = "62375"
ST_MARYS_CITY = "62400"
ST_PAUL_CITY = "62500"
STAFFORD_CITY = "67775"
STARK_CITY = "68025"
STERLING_CITY = "68200"
STOCKTON_CITY = "68350"
STRONG_CITY_CITY = "68650"
STUTTGART_CDP = "68750"
SUBLETTE_CITY = "68775"
SUMMERFIELD_CITY = "68950"
SUN_CITY_CITY = "69275"
SUSANK_CITY = "69525"
SYCAMORE_CDP = "69725"
SYLVAN_GROVE_CITY = "69775"
SYLVIA_CITY = "69800"
SYRACUSE_CITY = "69850"
TALMAGE_CDP = "69925"
TAMPA_CITY = "70000"
TECUMSEH_CDP = "70075"
TESCOTT_CITY = "70250"
THAYER_CITY = "70300"
THE_HIGHLANDS_CITY = "70357"
TIMKEN_CITY = "70625"
TIPTON_CITY = "70675"
TONGANOXIE_CITY = "70800"
TOPEKA_CITY = "71000"
TORONTO_CITY = "71050"
TOWANDA_CITY = "71125"
TRIBUNE_CITY = "71450"
TROY_CITY = "71575"
TURON_CITY = "71800"
TYRO_CITY = "71925"
UDALL_CITY = "71950"
ULYSSES_CITY = "71975"
UNIONTOWN_CITY = "72525"
URBANA_CDP = "72600"
UTICA_CITY = "72650"
VALLEY_CENTER_CITY = "73250"
VALLEY_FALLS_CITY = "73300"
VASSAR_CDP = "73450"
VERMILLION_CITY = "73525"
VICTORIA_CITY = "73775"
VINING_CITY = "73950"
VIOLA_CITY = "74025"
VIRGIL_CITY = "74075"
WAKEENEY_CITY = "74450"
WABAUNSEE_CDP = "74225"
WAKARUSA_CDP = "74425"
WAKEFIELD_CITY = "74525"
WALDO_CITY = "74575"
WALDRON_CITY = "74625"
WALLACE_CITY = "74750"
WALNUT_CITY = "74950"
WALTON_CITY = "75200"
WAMEGO_CITY = "75325"
WASHINGTON_CITY = "75825"
WATERVILLE_CITY = "75950"
WATHENA_CITY = "76000"
WAVERLY_CITY = "76050"
WAYSIDE_CDP = "76175"
WEBBER_CITY = "76250"
WEIR_CITY = "76350"
WELDA_CDP = "76425"
WELLINGTON_CITY = "76475"
WELLS_CDP = "76550"
WELLSVILLE_CITY = "76625"
WESKAN_CDP = "76675"
WEST_MINERAL_CITY = "77150"
WESTMORELAND_CITY = "77200"
WESTPHALIA_CITY = "77250"
WESTWOOD_HILLS_CITY = "77525"
WESTWOOD_CITY = "77500"
WETMORE_CITY = "77550"
WHEATON_CITY = "77700"
WHITE_CITY_CITY = "77825"
WHITE_CLOUD_CITY = "77850"
WHITEWATER_CITY = "78050"
WHITING_CITY = "78100"
WICHITA_CITY = "79000"
WILLARD_CITY = "79250"
WILLIAMSBURG_CITY = "79300"
WILLIAMSTOWN_CDP = "79375"
WILLIS_CITY = "79400"
WILLOWBROOK_CITY = "79425"
WILMORE_CITY = "79550"
WILROADS_GARDENS_CDP = "79625"
WILSEY_CITY = "79650"
WILSON_CITY = "79675"
WINCHESTER_CITY = "79800"
WINDOM_CITY = "79850"
WINFIELD_CITY = "79950"
WINONA_CITY = "80075"
WOODBINE_CITY = "80350"
WOODRUFF_CDP = "80400"
WOODSTON_CITY = "80450"
WRIGHT_CDP = "80500"
YALE_CDP = "80650"
YATES_CENTER_CITY = "80700"
YODER_CDP = "80750"
ZEANDALE_CDP = "80875"
ZENDA_CITY = "80925"
ZURICH_CITY = "81025"



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Illinois."""


ADAMS = "001"
ALEXANDER = "003"
BOND = "005"
BOONE = "007"
BROWN = "009"
BUREAU = "011"
CALHOUN = "013"
CARROLL = "015"
CASS = "017"
CHAMPAIGN = "019"
CHRISTIAN = "021"
CLARK = "023"
CLAY = "025"
CLINTON = "027"
COLES = "029"
COOK = "031"
CRAWFORD = "033"
CUMBERLAND = "035"
DE_WITT = "039"
DEKALB = "037"
DOUGLAS = "041"
DUPAGE = "043"
EDGAR = "045"
EDWARDS = "047"
EFFINGHAM = "049"
FAYETTE = "051"
FORD = "053"
FRANKLIN = "055"
FULTON = "057"
GALLATIN = "059"
GREENE = "061"
GRUNDY = "063"
HAMILTON = "065"
HANCOCK = "067"
HARDIN = "069"
HENDERSON = "071"
HENRY = "073"
IROQUOIS = "075"
JACKSON = "077"
JASPER = "079"
JEFFERSON = "081"
JERSEY = "083"
JO_DAVIESS = "085"
JOHNSON = "087"
KANE = "089"
KANKAKEE = "091"
KENDALL = "093"
KNOX = "095"
LASALLE = "099"
LAKE = "097"
LAWRENCE = "101"
LEE = "103"
LIVINGSTON = "105"
LOGAN = "107"
MACON = "115"
MACOUPIN = "117"
MADISON = "119"
MARION = "121"
MARSHALL = "123"
MASON = "125"
MASSAC = "127"
MCDONOUGH = "109"
MCHENRY = "111"
MCLEAN = "113"
MENARD = "129"
MERCER = "131"
MONROE = "133"
MONTGOMERY = "135"
MORGAN = "137"
MOULTRIE = "139"
OGLE = "141"
PEORIA = "143"
PERRY = "145"
PIATT = "147"
PIKE = "149"
POPE = "151"
PULASKI = "153"
PUTNAM = "155"
RANDOLPH = "157"
RICHLAND = "159"
ROCK_ISLAND = "161"
SALINE = "165"
SANGAMON = "167"
SCHUYLER = "169"
SCOTT = "171"
SHELBY = "173"
ST_CLAIR = "163"
STARK = "175"
STEPHENSON = "177"
TAZEWELL = "179"
UNION = "181"
VERMILION = "183"
WABASH = "185"
WARREN = "187"
WASHINGTON = "189"
WAYNE = "191"
WHITE = "193"
WHITESIDE = "195"
WILL = "197"
WILLIAMSON = "199"
WINNEBAGO = "201"
WOODFORD = "203"



# Copyright (c) 2023 Darren Erik Vengroff
"""Places in Arkansas."""


ACORN_CDP = "00190"
ADONA_CITY = "00250"
ALEXANDER_CITY = "00580"
ALICIA_TOWN = "00700"
ALIX_CDP = "00730"
ALLEENE_CDP = "00820"
ALLPORT_TOWN = "00940"
ALMA_CITY = "00970"
ALMYRA_TOWN = "01030"
ALPENA_TOWN = "01060"
ALPINE_CDP = "01090"
ALTHEIMER_CITY = "01150"
ALTUS_CITY = "01210"
AMAGON_TOWN = "01270"
AMITY_CITY = "01360"
ANTHONYVILLE_TOWN = "01457"
ANTOINE_TOWN = "01540"
APLIN_CDP = "01630"
APPLETON_CDP = "01660"
ARKADELPHIA_CITY = "01870"
ARKANSAS_CITY_CITY = "01990"
ARMOREL_CDP = "02230"
ASH_FLAT_CITY = "02470"
ASHDOWN_CITY = "02380"
ATKINS_CITY = "02590"
AUBREY_TOWN = "02680"
AUGUSTA_CITY = "02740"
AUSTIN_CITY = "02860"
AVILLA_CDP = "03010"
AVOCA_TOWN = "03040"
BALD_KNOB_CITY = "03280"
BANKS_TOWN = "03430"
BARLING_CITY = "03640"
BASSETT_TOWN = "03940"
BATAVIA_CDP = "03970"
BATESVILLE_CITY = "04030"
BAUXITE_TOWN = "04090"
BAY_CITY = "04180"
BEARDEN_CITY = "04420"
BEAVER_TOWN = "04540"
BEE_BRANCH_CDP = "04630"
BEEBE_CITY = "04600"
BEEDEVILLE_TOWN = "04720"
BEIRNE_CDP = "04750"
BELLA_VISTA_CITY = "04840"
BELLEFONTE_TOWN = "04900"
BELLEVILLE_CITY = "04960"
BEN_LOMOND_TOWN = "05260"
BENTON_CITY = "05290"
BENTONVILLE_CITY = "05320"
BERGMAN_TOWN = "05440"
BERRYVILLE_CITY = "05560"
BETHESDA_CDP = "05770"
BIG_FLAT_TOWN = "05980"
BIGELOW_TOWN = "05920"
BIGGERS_TOWN = "06040"
BIRDSONG_TOWN = "06340"
BISMARCK_CDP = "06460"
BLACK_OAK_TOWN = "06610"
BLACK_ROCK_CITY = "06700"
BLACK_SPRINGS_TOWN = "06730"
BLEVINS_CITY = "07030"
BLUE_EYE_TOWN = "07150"
BLUE_MOUNTAIN_TOWN = "07210"
BLUFF_CITY_TOWN = "07270"
BLYTHEVILLE_CITY = "07330"
BOARD_CAMP_CDP = "07420"
BODCAW_TOWN = "07450"
BOLES_CDP = "07510"
BONANZA_CITY = "07540"
BONO_CITY = "07630"
BOONEVILLE_CITY = "07720"
BOWMAN_CDP = "08020"
BRADFORD_CITY = "08260"
BRADLEY_CITY = "08290"
BRANCH_CITY = "08440"
BRIARCLIFF_CITY = "08665"
BRINKLEY_CITY = "08950"
BROOKLAND_CITY = "09100"
BRYANT_CITY = "09460"
BUCKNER_CITY = "09550"
BUFFALO_CITY_CDP = "09700"
BULL_SHOALS_CITY = "09790"
BURDETTE_TOWN = "09880"
CABOT_CITY = "10300"
CADDO_GAP_CDP = "10330"
CADDO_VALLEY_TOWN = "10360"
CALDWELL_TOWN = "10450"
CALE_TOWN = "10480"
CALICO_ROCK_CITY = "10570"
CALION_CITY = "10600"
CAMDEN_CITY = "10720"
CAMMACK_VILLAGE_CITY = "10780"
CAMPBELL_STATION_CITY = "10900"
CANEHILL_CDP = "11080"
CARAWAY_CITY = "11410"
CARLISLE_CITY = "11500"
CARTHAGE_CITY = "11830"
CASA_TOWN = "11890"
CASH_TOWN = "11920"
CAULKSVILLE_TOWN = "12190"
CAVE_CITY_CITY = "12280"
CAVE_SPRINGS_CITY = "12340"
CEDARVILLE_CITY = "12520"
CENTER_POINT_CDP = "12700"
CENTER_RIDGE_CDP = "12790"
CENTERTON_CITY = "12820"
CENTERVILLE_CDP = "12940"
CENTRAL_CITY_TOWN = "13120"
CHARLESTON_CITY = "13300"
CHEROKEE_CITY_CDP = "13420"
CHEROKEE_VILLAGE_CITY = "13450"
CHERRY_VALLEY_CITY = "13540"
CHESTER_TOWN = "13570"
CHIDESTER_CITY = "13750"
CINCINNATI_CDP = "13960"
CLARENDON_CITY = "13990"
CLARKEDALE_CITY = "14050"
CLARKSVILLE_CITY = "14140"
CLINTON_CITY = "14260"
COAL_HILL_CITY = "14500"
COLLEGE_STATION_CDP = "14860"
COLLINS_CDP = "14920"
COLT_CITY = "14950"
CONCORD_TOWN = "15100"
CONWAY_CITY = "15190"
CORINTH_TOWN = "15310"
CORNING_CITY = "15460"
COTTER_CITY = "15490"
COTTON_PLANT_CITY = "15550"
COVE_TOWN = "15700"
COY_TOWN = "15790"
CRAWFORDSVILLE_TOWN = "15940"
CROSSETT_CITY = "16240"
CRYSTAL_SPRINGS_CDP = "16690"
CUSHMAN_CITY = "16930"
DAISY_TOWN = "17140"
DAMASCUS_TOWN = "17290"
DANVILLE_CITY = "17320"
DARDANELLE_CITY = "17380"
DATTO_TOWN = "17410"
DE_QUEEN_CITY = "18490"
DE_VALLS_BLUFF_CITY = "18640"
DEWITT_CITY = "18790"
DECATUR_CITY = "17740"
DEER_CDP = "17860"
DELAPLAINE_TOWN = "18010"
DELIGHT_CITY = "18100"
DELL_TOWN = "18160"
DENNARD_CDP = "18340"
DENNING_TOWN = "18370"
DERMOTT_CITY = "18520"
DES_ARC_CITY = "18550"
DESHA_CDP = "18580"
DIAMOND_CITY_CITY = "18850"
DIAZ_CITY = "18940"
DIERKS_CITY = "18970"
DONALDSON_TOWN = "19450"
DORA_CDP = "19540"
DOVER_CITY = "19600"
DRASCO_CDP = "19720"
DUMAS_CITY = "19990"
DYER_CITY = "20200"
DYESS_TOWN = "20230"
EARLE_CITY = "20320"
EAST_CAMDEN_TOWN = "20440"
EAST_END_CDP = "20470"
EDGEMONT_CDP = "20800"
EDMONDSON_TOWN = "20830"
EGYPT_TOWN = "20920"
EL_DORADO_CITY = "21070"
EL_PASO_CDP = "21520"
ELAINE_CITY = "20950"
ELKINS_CITY = "21190"
ELM_SPRINGS_CITY = "21430"
EMERSON_TOWN = "21580"
EMMET_CITY = "21610"
ENGLAND_CITY = "21730"
ENOLA_TOWN = "21820"
ETOWAH_TOWN = "22120"
EUDORA_CITY = "22180"
EUREKA_SPRINGS_CITY = "22240"
EVANSVILLE_CDP = "22300"
EVENING_SHADE_CITY = "22360"
EVERTON_TOWN = "22450"
FAIR_OAKS_CDP = "22720"
FAIRFIELD_BAY_CITY = "22660"
FARGO_TOWN = "23140"
FARMINGTON_CITY = "23170"
FAYETTEVILLE_CITY = "23290"
FELSENTHAL_TOWN = "23320"
FIFTY_SIX_CITY = "23680"
FISHER_CITY = "23800"
FLIPPIN_CITY = "24010"
FLORAL_CDP = "24070"
FLOYD_CDP = "24130"
FORDYCE_CITY = "24220"
FOREMAN_CITY = "24250"
FORREST_CITY_CITY = "24430"
FORT_SMITH_CITY = "24550"
FOUKE_CITY = "24640"
FOUNTAIN_HILL_TOWN = "24670"
FOUNTAIN_LAKE_TOWN = "24700"
FOURCHE_TOWN = "24760"
FOX_CDP = "24910"
FRANKLIN_TOWN = "25030"
FREDONIA_BISCOE_TOWN = "25060"
FRIENDSHIP_TOWN = "25180"
FULTON_TOWN = "25360"
GAMALIEL_CDP = "25630"
GARFIELD_TOWN = "25750"
GARLAND_TOWN = "25780"
GARNER_TOWN = "25900"
GASSVILLE_CITY = "26050"
GATEWAY_TOWN = "26110"
GENOA_CDP = "26260"
GENTRY_CITY = "26290"
GEORGETOWN_TOWN = "26440"
GIBSON_CDP = "26710"
GILBERT_TOWN = "26800"
GILLETT_CITY = "26980"
GILLHAM_TOWN = "27010"
GILMORE_CITY = "27040"
GLENWOOD_CITY = "27310"
GOODWIN_CDP = "27610"
GOSHEN_TOWN = "27670"
GOSNELL_CITY = "27700"
GOULD_CITY = "27730"
GRADY_CITY = "27790"
GRANNIS_CITY = "27970"
GRAVETTE_CITY = "28360"
GREEN_FOREST_CITY = "28600"
GREENBRIER_CITY = "28510"
GREENLAND_CITY = "28660"
GREENWAY_CITY = "28720"
GREENWOOD_CITY = "28780"
GREERS_FERRY_CITY = "28810"
GREGORY_CDP = "28840"
GRIFFITHVILLE_TOWN = "28990"
GRUBBS_CITY = "29020"
GUION_TOWN = "29080"
GUM_SPRINGS_TOWN = "29140"
GURDON_CITY = "29200"
GUY_CITY = "29230"
HACKETT_CITY = "29290"
HAGARVILLE_CDP = "29320"
HALLEY_CDP = "29380"
HAMBURG_CITY = "29500"
HAMPTON_CITY = "29650"
HARDY_CITY = "29920"
HARRELL_TOWN = "30310"
HARRISBURG_CITY = "30400"
HARRISON_CITY = "30460"
HARTFORD_CITY = "30490"
HARTMAN_CITY = "30520"
HASKELL_CITY = "30640"
HATFIELD_TOWN = "30700"
HATTIEVILLE_CDP = "30730"
HAVANA_CITY = "30790"
HAYNES_TOWN = "30880"
HAZEN_CITY = "30940"
HEBER_SPRINGS_CITY = "31090"
HECTOR_TOWN = "31150"
HELENA_WEST_HELENA_CITY = "31190"
HENDERSON_CDP = "31270"
HENSLEY_CDP = "31390"
HERMITAGE_CITY = "31540"
HICKORY_RIDGE_CITY = "31900"
HIGDEN_TOWN = "32080"
HIGGINSON_TOWN = "32140"
HIGHFILL_TOWN = "32200"
HIGHLAND_CITY = "32257"
HINDSVILLE_TOWN = "32470"
HOLIDAY_ISLAND_TOWN = "32650"
HOLLAND_CITY = "32710"
HOLLY_GROVE_CITY = "32800"
HOPE_CITY = "33190"
HORATIO_CITY = "33310"
HORSESHOE_BEND_CITY = "33370"
HORSESHOE_LAKE_TOWN = "33375"
HOT_SPRINGS_VILLAGE_CDP = "33482"
HOT_SPRINGS_CITY = "33400"
HOUSTON_TOWN = "33490"
HOXIE_CITY = "33580"
HUGHES_CITY = "33760"
HUMNOKE_CITY = "33820"
HUMPHREY_CITY = "33850"
HUNTER_TOWN = "33910"
HUNTINGTON_CITY = "33940"
HUNTSVILLE_CITY = "33970"
HUTTIG_CITY = "34090"
IMBODEN_TOWN = "34150"
INDIAN_BAY_CDP = "34220"
IVAN_CDP = "34540"
JACKSONPORT_TOWN = "34720"
JACKSONVILLE_CITY = "34750"
JASPER_CITY = "34900"
JENNETTE_TOWN = "35040"
JERICHO_TOWN = "35140"
JEROME_CDP = "35170"
JERUSALEM_CDP = "35260"
JOHNSON_CITY = "35500"
JOINER_CITY = "35650"
JONES_MILLS_CDP = "35770"
JONESBORO_CITY = "35710"
JUDSONIA_CITY = "36040"
JUNCTION_CITY_CITY = "36130"
KEISER_CITY = "36310"
KENSETT_CITY = "36430"
KEO_TOWN = "36550"
KIBLER_CITY = "36670"
KINGSLAND_CITY = "36880"
KINGSTON_CDP = "36910"
KIRBY_CDP = "37060"
KNOBEL_CITY = "37240"
KNOXVILLE_TOWN = "37300"
LAGRANGE_TOWN = "37720"
LACEY_CDP = "37450"
LAFE_TOWN = "37660"
LAKE_CITY_CITY = "37780"
LAKE_HAMILTON_CDP = "37930"
LAKE_VIEW_CITY = "38110"
LAKE_VILLAGE_CITY = "38170"
LAKEVIEW_CITY = "38050"
LAMAR_CITY = "38290"
LANDMARK_CDP = "38540"
LAVACA_CITY = "38890"
LAWSON_CDP = "38950"
LEACHVILLE_CITY = "39010"
LEAD_HILL_TOWN = "39040"
LEOLA_TOWN = "39310"
LEPANTO_CITY = "39370"
LESLIE_CITY = "39460"
LETONA_TOWN = "39520"
LEWISVILLE_CITY = "39640"
LEXA_TOWN = "39670"
LINCOLN_CITY = "39970"
LITTLE_FLOCK_CITY = "40120"
LITTLE_ROCK_CITY = "41000"
LOCKESBURG_CITY = "41090"
LONDON_CITY = "41270"
LONOKE_CITY = "41420"
LONSDALE_TOWN = "41450"
LOST_BRIDGE_VILLAGE_CDP = "41555"
LOUANN_TOWN = "41630"
LOWELL_CITY = "41720"
LUXORA_CITY = "42170"
LYNN_TOWN = "42260"
MADISON_CITY = "43280"
MAGAZINE_CITY = "43310"
MAGNESS_TOWN = "43370"
MAGNET_COVE_CDP = "43430"
MAGNOLIA_CITY = "43460"
MALVERN_CITY = "43610"
MAMMOTH_SPRING_CITY = "43670"
MANILA_CITY = "43820"
MANSFIELD_CITY = "43880"
MARIANNA_CITY = "44120"
MARIE_TOWN = "44150"
MARION_CITY = "44180"
MARKED_TREE_CITY = "44210"
MARMADUKE_CITY = "44240"
MARSHALL_CITY = "44300"
MARVELL_CITY = "44420"
MAUMELLE_CITY = "44600"
MAYFLOWER_CITY = "44750"
MAYNARD_TOWN = "44780"
MAYSVILLE_CDP = "44810"
MCALMONT_CDP = "42350"
MCCASKILL_TOWN = "42410"
MCCRORY_CITY = "42500"
MCDOUGAL_TOWN = "42560"
MCGEHEE_CITY = "42770"
MCNAB_TOWN = "43070"
MCNEIL_CITY = "43100"
MCRAE_CITY = "43220"
MELBOURNE_CITY = "45080"
MELLWOOD_CDP = "45110"
MENA_CITY = "45170"
MENIFEE_TOWN = "45200"
MIDLAND_TOWN = "45500"
MIDWAY_CDP = "45530"
MIDWAY_TOWN = "45560"
MINERAL_SPRINGS_CITY = "46040"
MINTURN_TOWN = "46100"
MITCHELLVILLE_CITY = "46190"
MONETTE_CITY = "46400"
MONROE_CDP = "46460"
MONTICELLO_CITY = "46580"
MONTROSE_CITY = "46670"
MOOREFIELD_TOWN = "46730"
MORO_TOWN = "46910"
MORRILTON_CITY = "46970"
MORRISON_BLUFF_TOWN = "47030"
MORROW_CDP = "47090"
MOUNT_HOLLY_CDP = "47660"
MOUNT_IDA_CITY = "47690"
MOUNT_JUDEA_CDP = "47720"
MOUNT_OLIVE_CDP = "47810"
MOUNT_PLEASANT_TOWN = "47900"
MOUNT_VERNON_TOWN = "48020"
MOUNTAIN_HOME_CITY = "47390"
MOUNTAIN_PINE_CITY = "47450"
MOUNTAIN_VIEW_CITY = "47540"
MOUNTAINBURG_CITY = "47300"
MULBERRY_CITY = "48200"
MURFREESBORO_CITY = "48290"
NASHVILLE_CITY = "48560"
NATURAL_STEPS_CDP = "48680"
NEW_BLAINE_CDP = "49070"
NEW_EDINBURG_CDP = "49220"
NEWARK_CITY = "49010"
NEWHOPE_CDP = "49370"
NEWPORT_CITY = "49580"
NIMMONS_TOWN = "49820"
NORFORK_CITY = "50000"
NORMAN_TOWN = "50030"
NORPHLET_CITY = "50060"
NORTH_CROSSETT_CDP = "50240"
NORTH_LITTLE_ROCK_CITY = "50450"
OKEAN_TOWN = "51500"
OAK_GROVE_CDP = "50990"
OAK_GROVE_HEIGHTS_TOWN = "51060"
OAK_GROVE_TOWN = "50810"
OAKHAVEN_TOWN = "51080"
OAKLAND_CDP = "51110"
OARK_CDP = "51230"
ODEN_TOWN = "51290"
OGDEN_CITY = "51350"
OIL_TROUGH_TOWN = "51410"
OKOLONA_TOWN = "51530"
OLA_CITY = "51560"
OMAHA_TOWN = "52160"
ONEIDA_CDP = "52280"
OPPELO_CITY = "52430"
OSCEOLA_CITY = "52580"
OXFORD_CITY = "52880"
OZAN_TOWN = "52940"
OZARK_ACRES_CDP = "53000"
OZARK_CITY = "52970"
OZONE_CDP = "53060"
PALESTINE_CITY = "53150"
PANGBURN_CITY = "53240"
PARAGOULD_CITY = "53390"
PARIS_CITY = "53480"
PARKDALE_CITY = "53510"
PARKIN_CITY = "53600"
PATMOS_TOWN = "53900"
PATTERSON_CITY = "53990"
PAYNEWAY_CDP = "54110"
PEA_RIDGE_CITY = "54200"
PEACH_ORCHARD_CITY = "54140"
PEARCY_CDP = "54170"
PENCIL_BLUFF_CDP = "54410"
PERLA_TOWN = "54620"
PERRY_TOWN = "54650"
PERRYTOWN_TOWN = "54680"
PERRYVILLE_CITY = "54710"
PIGGOTT_CITY = "55130"
PINDALL_TOWN = "55280"
PINE_BLUFF_CITY = "55310"
PINEVILLE_TOWN = "55580"
PINEY_CDP = "55610"
PLAINVIEW_CITY = "55970"
PLEASANT_GROVE_CDP = "56060"
PLEASANT_PLAINS_TOWN = "56270"
PLUMERVILLE_CITY = "56480"
POCAHONTAS_CITY = "56540"
POLLARD_CITY = "56600"
PONCA_CDP = "56630"
POPLAR_GROVE_CDP = "56690"
PORTIA_TOWN = "56720"
PORTLAND_CITY = "56750"
POTTSVILLE_CITY = "56990"
POWHATAN_TOWN = "57050"
POYEN_TOWN = "57080"
PRAIRIE_CREEK_CDP = "57125"
PRAIRIE_GROVE_CITY = "57170"
PRATTSVILLE_TOWN = "57230"
PRESCOTT_CITY = "57260"
PRINCETON_CDP = "57530"
PYATT_TOWN = "57890"
QUITMAN_CITY = "57950"
RATCLIFF_CITY = "58160"
RAVENDEN_SPRINGS_TOWN = "58280"
RAVENDEN_TOWN = "58250"
READER_CDP = "58400"
RECTOR_CITY = "58490"
REDFIELD_CITY = "58580"
REED_TOWN = "58880"
REYNO_CITY = "59180"
RISON_CITY = "59480"
RIVERVALE_CDP = "59630"
ROCKPORT_CITY = "60020"
ROCKWELL_CDP = "60110"
ROE_TOWN = "60380"
ROGERS_CITY = "60410"
ROLAND_CDP = "60470"
RONDO_TOWN = "60590"
ROSE_BUD_TOWN = "60770"
ROSSTON_TOWN = "61040"
ROVER_CDP = "61190"
RUDY_TOWN = "61460"
RUSSELL_TOWN = "61640"
RUSSELLVILLE_CITY = "61670"
RYE_CDP = "61730"
SALADO_CDP = "62120"
SALEM_CDP = "62210"
SALEM_CITY = "62150"
SALESVILLE_CITY = "62240"
SARATOGA_CDP = "62630"
SARDIS_CDP = "62660"
SCOTT_CDP = "62900"
SCRANTON_CITY = "62960"
SEARCY_CITY = "63020"
SEDGWICK_TOWN = "63110"
SHANNON_HILLS_CITY = "63470"
SHERIDAN_CITY = "63710"
SHERRILL_TOWN = "63740"
SHERWOOD_CITY = "63800"
SHIRLEY_TOWN = "63980"
SIDNEY_TOWN = "64280"
SILOAM_SPRINGS_CITY = "64370"
SMACKOVER_CITY = "64730"
SMITHVILLE_TOWN = "64910"
SOUTH_LEAD_HILL_TOWN = "65480"
SOUTHSIDE_CITY = "65630"
SPARKMAN_CITY = "65900"
SPRINGDALE_CITY = "66080"
SPRINGFIELD_CDP = "66110"
SPRINGTOWN_TOWN = "66200"
ST_CHARLES_TOWN = "61940"
ST_FRANCIS_CITY = "61970"
ST_JOE_TOWN = "62030"
ST_PAUL_TOWN = "62060"
STAMPS_CITY = "66320"
STAR_CITY_CITY = "66440"
STAVES_CDP = "66800"
STEPHENS_CITY = "66860"
STRAWBERRY_TOWN = "67250"
STRONG_CITY = "67370"
STUTTGART_CITY = "67490"
SUBIACO_TOWN = "67520"
SUCCESS_TOWN = "67550"
SULPHUR_ROCK_TOWN = "67730"
SULPHUR_SPRINGS_CDP = "67790"
SULPHUR_SPRINGS_CITY = "67760"
SUMMERS_CDP = "67880"
SUMMIT_CITY = "67940"
SUNSET_TOWN = "68060"
SWEET_HOME_CDP = "68330"
SWIFTON_CITY = "68360"
TAYLOR_CITY = "68660"
TEXARKANA_CITY = "68810"
THORNTON_CITY = "69050"
TILLAR_CITY = "69230"
TINSMAN_TOWN = "69350"
TOLLETTE_TOWN = "69500"
TONTITOWN_CITY = "69740"
TRASKWOOD_CITY = "69830"
TRUMANN_CITY = "70010"
TUCKER_CDP = "70070"
TUCKERMAN_CITY = "70100"
TULL_TOWN = "70220"
TUMBLING_SHOALS_CDP = "70310"
TUPELO_TOWN = "70340"
TURRELL_CITY = "70430"
TWIN_GROVES_TOWN = "70540"
TYRONZA_CITY = "70700"
ULM_TOWN = "70760"
UNIONTOWN_CDP = "71060"
URBANA_CDP = "71180"
VALLEY_SPRINGS_TOWN = "71390"
VAN_BUREN_CITY = "71480"
VANDERVOORT_TOWN = "71510"
VANNDALE_CDP = "71570"
VICTORIA_TOWN = "71900"
VILONIA_CITY = "71960"
VIOLA_TOWN = "72140"
VIOLET_HILL_CDP = "72170"
WABBASEKA_TOWN = "72230"
WALCOTT_CDP = "72290"
WALDENBURG_TOWN = "72320"
WALDO_CITY = "72350"
WALDRON_CITY = "72380"
WALNUT_RIDGE_CITY = "72890"
WARD_CITY = "73130"
WARM_SPRINGS_CDP = "73250"
WARREN_CITY = "73310"
WASHINGTON_CITY = "73370"
WATSON_CITY = "73550"
WAYTON_CDP = "73730"
WEINER_CITY = "73940"
WELDON_TOWN = "74000"
WESLEY_CDP = "74090"
WEST_CROSSETT_CDP = "74270"
WEST_FORK_CITY = "74360"
WEST_MEMPHIS_CITY = "74540"
WEST_POINT_TOWN = "74660"
WESTERN_GROVE_TOWN = "74330"
WHEATLEY_CITY = "74840"
WHELEN_SPRINGS_TOWN = "74930"
WHITE_HALL_CITY = "75170"
WICKES_TOWN = "75500"
WIDENER_TOWN = "75560"
WIEDERKEHR_VILLAGE_CITY = "75570"
WILBURN_CDP = "75590"
WILLIFORD_TOWN = "75740"
WILLISVILLE_TOWN = "75770"
WILMAR_CITY = "75860"
WILMOT_CITY = "75890"
WILSON_CITY = "75920"
WILTON_CITY = "75980"
WINCHESTER_TOWN = "76010"
WINSLOW_CITY = "76190"
WINTHROP_CITY = "76250"
WITTS_SPRINGS_CDP = "76460"
WOODLAWN_CDP = "76690"
WOODSON_CDP = "76730"
WOOSTER_TOWN = "76820"
WRIGHTSVILLE_CITY = "76970"
WYNNE_CITY = "77090"
YARBOROUGH_LANDING_CDP = "77203"
YELLVILLE_CITY = "77330"
ZINC_TOWN = "77600"



# Copyright (c) 2023 Darren Erik Vengroff
"""
Details of how U.S. Census shapefiles and their contents are named.

Also includes utilities for managing downloaded shapefiles and
for water clipping.
"""

from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Callable, Tuple, Optional, List, Generator, Union
from logging import getLogger

import geopandas as gpd
import pandas as pd

from censusdis import CensusApiException, maps as cmap
from censusdis.impl.geometry import drop_slivers_from_gdf
from censusdis.impl.varsource.base import VintageType


logger = getLogger(__name__)


_GEO_QUERY_FROM_DATA_QUERY_INNER_GEO: Dict[
    str,
    Callable[[int], Tuple[Optional[str], str, List[str], List[str]]],
] = {
    # innermost geo: ( shapefile_scope, shapefile_geo_name, df_on, gdf_on )
    "us": lambda year: ("us", "nation", ["US"], ["US"]),
    "region": lambda year: ("us", "region", ["REGION"], ["REGIONCE"]),
    "division": lambda year: ("us", "division", ["DIVISION"], ["DIVISIONCE"]),
    "combined statistical area": lambda year: (
        "us",
        "csa",
        ["COMBINED_STATISTICAL_AREA"],
        ["CSAFP"],
    ),
    "metropolitan statistical area/micropolitan statistical area": lambda year: (
        "us",
        "cbsa",
        ["METROPOLITAN_STATISTICAL_AREA_MICROPOLITAN_STATISTICAL_AREA"],
        ["CBSAFP"],
    ),
    "state": lambda year: ("us", "state", ["STATE"], ["STATEFP"]),
    "consolidated city": lambda year: (
        "us",
        "concity",
        ["STATE", "CONSOLIDATED_CITY"],
        ["STATEFP", "CONCTYFP"],
    ),
    "county": lambda year: (
        "us",
        "county",
        ["STATE", "COUNTY"],
        ["STATEFP", "COUNTYFP"],
    ),
    "public use microdata area": lambda year: (
        "us",
        "puma10" if year < 2020 else "puma20",
        ["STATE", "PUBLIC_USE_MICRODATA_AREA"],
        ["STATEFP", "PUMACE"] if year < 2020 else ["STATEFP20", "PUMACE20"],
    ),
    "congressional district": lambda year: (
        "us",
        _congressional_district_from_year(year),
        ["STATE", "CONGRESSIONAL_DISTRICT"],
        ["STATEFP", f"{_congressional_district_from_year(year).upper()}FP"],
    ),
    "zip code tabulation area": lambda year: (
        "us",
        "zcta520" if year >= 2020 else "zcta510",
        ["ZIP_CODE_TABULATION_AREA"],
        ["ZCTA5CE10" if year < 2020 else "ZCTA5CE20" if year == 2020 else "ZCTA5CE"],
    ),
    "american indian area/alaska native area/hawaiian home land": lambda year: (
        "us",
        "aiannh",
        ["AMERICAN_INDIAN_AREA_ALASKA_NATIVE_AREA_HAWAIIAN_HOME_LAND"],
        ["AIANNHCE"],
    ),
    "new england city and town area": lambda year: (
        "us",
        "necta",
        ["NEW_ENGLAND_CITY_AND_TOWN_AREA"],
        ["NECTAFP"],
    ),
    # For these, the shapefiles are at the state level, so `None`
    # indicates that we have to fill it in based on the geometry
    # being queried.
    "county subdivision": lambda year: (
        None,
        "cousub",
        ["STATE", "COUNTY_SUBDIVISION"],
        ["STATEFP", "COUSUB" if year == 2010 else "COUSUBFP"],
    ),
    "place": lambda year: (None, "place", ["STATE", "PLACE"], ["STATEFP", "PLACEFP"]),
    "tract": lambda year: (
        None,
        "tract",
        ["STATE", "COUNTY", "TRACT"],
        ["STATEFP", "COUNTYFP", "TRACTCE"],
    ),
    "block group": lambda year: (
        None,
        "bg",
        ["STATE", "COUNTY", "TRACT", "BLOCK_GROUP"],
        ["STATEFP", "COUNTYFP", "TRACTCE", "BLKGRPCE"],
    ),
    "block": lambda year: (
        None,
        "tabblock",
        ["STATE", "COUNTY", "TRACT", "BLOCK"],
        ["STATEFP", "COUNTYFP", "TRACTCE", "BLOCKCE"],
    ),
    "school district (unified)": lambda year: (
        None,
        "unsd",
        ["STATE", "SCHOOL_DISTRICT_UNIFIED"],
        ["STATEFP", "UNSDLEA"],
    ),
    "school district (elementary)": lambda year: (
        # CB files from 2016 on exist with the "elsd" name and "ELSDLEA" column.
        # Earlier tiger files use the name "sda" and "ESDLEA" column.
        None,
        "elsd" if year >= 2016 else "sde",
        ["STATE", "SCHOOL_DISTRICT_ELEMENTARY"],
        ["STATEFP", "ELSDLEA" if year >= 2016 else "ESDLEA"],
    ),
    "school district (secondary)": lambda year: (
        # Similar scenario to school district (secondary).
        None,
        "scsd" if year >= 2016 else "sde",
        ["STATE", "SCHOOL_DISTRICT_SECONDARY"],
        ["STATEFP", "SCSDLEA" if year >= 2016 else "SSDLEA"],
    ),
    "state legislative district (upper chamber)": lambda year: (
        None,
        "sldu",
        ["STATE", "STATE_LEGISLATIVE_DISTRICT_UPPER_CHAMBER"],
        ["STATEFP", "SLDUST"],
    ),
    "state legislative district (lower chamber)": lambda year: (
        None,
        "sldl",
        ["STATE", "STATE_LEGISLATIVE_DISTRICT_LOWER_CHAMBER"],
        ["STATEFP", "SLDLST"],
    ),
    "voting district": lambda year: (
        None,
        "vtd",
        ["STATE", "COUNTY", "VOTING_DISTRICT"],
        (
            ["STATEFP20", "COUNTYFP20", "VTDST20"]
            if year >= 2020
            else ["STATEFP10", "COUNTYFP10", "VTDST10"]
        ),
    ),
    # This one could be a little dangerous if subminor civil
    # divisions exist in states and are not mapped as subbarios.
    # It appears the only example of this is the ESTATE in the
    # USVI. So we are going to punt for the moment and deal with
    # it at some future time, probably by adding an arg to the
    # lambdas to take in the bound params so we can look at them.
    #
    # Noted in https://github.com/vengroff/censusdis/issues/223.
    "subminor civil division": lambda year: (
        None,
        "subbarrio",
        ["STATE", "COUNTY", "COUNTY_SUBDIVISION", "SUBMINOR_CIVIL_DIVISION"],
        ["STATEFP", "COUNTYFP", "COUSUBFP", "SUBMCDFP"],
    ),
    "alaska native regional corporation": lambda year: (
        None,
        "anrc",
        ["ALASKA_NATIVE_REGIONAL_CORPORATION"],
        ["ANRCFP"],
    ),
}
"""
Helper map for the _with_geometry case.

A map from the innermost level of a geometry specification
to the arguments we need to pass to `get_cb_shapefile`
to get the right shapefile for the geography and the columns
we need to join the data and shapefile on.

The values are functions of the year. Once we locate one, we
call it with the year as an argument to get the final value.
This is necessary for e.g. congressional districts, which have
names that change every two years. It also turns out to be very
helpful in dealing with little quirks in the data like changing
the name of the COUSUB/COSUBFP column in just one year.

We should add everything in
https://www.census.gov/geographies/mapping-files/time-series/geo/cartographic-boundary.html
to this map.
"""


def geo_query_from_data_query_inner_geo(
    year: int, geo_level: str
) -> Tuple[Optional[str], str, List[str], List[str]]:
    """
    Map lookup and call the for the give year to produce the result.

    Parameters
    ----------
    year
        The year to pass to callable values.
    geo_level
        The key to look up.

    Returns
    -------
        A tuple of results.
    """
    if geo_level not in _GEO_QUERY_FROM_DATA_QUERY_INNER_GEO:
        raise CensusApiException(
            "The with_geometry=True flag is only allowed if the "
            f"geometry for the data to be loaded ('{geo_level}') is one of "
            f"{list(_GEO_QUERY_FROM_DATA_QUERY_INNER_GEO.keys())}."
        )

    return _GEO_QUERY_FROM_DATA_QUERY_INNER_GEO[geo_level](year)


def _geo_query_from_data_query_inner_geo_items(
    year: int,
) -> Generator[Tuple[Optional[str], str, List[str], List[str]], None, None]:
    """Generate the items in `GEO_QUERY_FROM_DATA_QUERY_INNER_GEO`."""
    for geo_level in _GEO_QUERY_FROM_DATA_QUERY_INNER_GEO:
        yield geo_level, geo_query_from_data_query_inner_geo(year, geo_level)


def _congressional_district_from_year(year: int) -> str:
    """Construct the short form of the congressional district used in a given year."""
    # See the files in https://www2.census.gov/geo/tiger/GENZ2020/shp/
    # and similar. The interesting ones are of the form
    #
    # cb_20YY_us_cdCCC_500k.zip
    #
    # where YY is the year and CCC is the congressional district
    # used.
    #
    # The mappings are not exactly at two year intervals as we would expect.
    if year == 2020 or year == 2021:
        # For some reason they did not update to cd117 for these years. Pandemic?
        return "cd116"

    # Regular year rule.
    congress = 104 + (year - 1994) // 2
    return f"cd{congress}"


def infer_geo_level(year: Optional[int], df_data: pd.DataFrame) -> str:
    """
    Infer the geography level based on columns names.

    Parameters
    ----------
    year
        The vintage of the data. `None` to infer from the data.
    df_data
        A dataframe of variables with one or more columns that
        can be used to infer what geometry level the rows represent.

        For example, if the column `"STATE"` exists, we could infer that
        the data in on a state by state basis. But if there are
        columns for both `"STATE"` and `"COUNTY"`, the data is probably
        at the county level.

        If, on the other hand, there is a `"COUNTY" column but not a
        `"STATE"` column, then there is some ambiguity. The data
        probably corresponds to counties, but the same county ID can
        exist in multiple states, so we will raise a
        :py:class:`~CensusApiException` with an error message expalining
        the situation.

        If there is no match, we will also raise an exception. Again we
        do this, rather than for example, returning `None`, so that we
        can provide an informative error message about the likely cause
        and what to do about it.

        This function is not often called directly, but rather from
        :py:func:`~add_inferred_geography`, which infers the geography
        level and then adds a `geometry` column containing the appropriate
        geography for each row.

    Returns
    -------
        The name of the geography level.
    """
    match_key = None
    match_on_len = 0
    partial_match_keys = []

    for k, (_, _, df_on, _) in _geo_query_from_data_query_inner_geo_items(year):
        if all(col in df_data.columns for col in df_on):
            # Full match. We want the longest full match
            # we find.
            if match_key is None or len(df_on) > match_on_len:
                match_key = k
        elif df_on[-1] in df_data.columns:
            # Partial match. This could result in us
            # not getting what we expect. Like if we
            # have STATE and TRACT, but not COUNTY, we will
            # get a partial match on [STATE, COUNTY. TRACT]
            # and a full match on [STATE]. We probably did
            # not mean to match [STATE], we just lost the
            # COUNTY column somewhere along the way.
            partial_match_keys.append(k)

    if match_key is None:
        raise CensusApiException(
            f"Unable to infer geometry. Was not able to locate any of the "
            "known sets of columns "
            f"{tuple(df_on for _, (_, _, df_on, _) in _geo_query_from_data_query_inner_geo_items(year))} "
            f"in the columns {list(df_data.columns)}."
        )

    if partial_match_keys:
        raise CensusApiException(
            f"Unable to infer geometry. Geometry matched {match_key} on columns "
            f"{geo_query_from_data_query_inner_geo(year, match_key)[2]} "
            "but also partially matched one or more candidates "
            f"{tuple(geo_query_from_data_query_inner_geo(year, k)[2] for k in partial_match_keys)}. "
            "Partial matches are usually unintended. Either add columns to allow a "
            "full match or rename some columns to prevent the undesired partial match."
        )

    return match_key


def add_geography(
    df_data: pd.DataFrame,
    year: Optional[VintageType],
    shapefile_scope: str,
    geo_level: str,
    with_geometry_columns: bool = False,
    tiger_shapefiles_only: bool = False,
) -> Union[pd.DataFrame, gpd.GeoDataFrame]:
    """
    Add geography to data.

    Parameters
    ----------
    df_data
        The data we downloaded from the census API
    year
        The year for which to fetch geometries. We need this
        because they change over time. If `None`, look for a
        `'YEAR'` column in `df_data` and possibly add different
        geometries for different years as needed.
    shapefile_scope
        The scope of the shapefile. This is typically either a state
        such as `NJ` or the string `"us"`.
    geo_level
        The geography level we want to add.
    with_geometry_columns
        If `True` keep all the additional columns that come with shapefiles
        downloaded to get geometry information.
    tiger_shapefiles_only
        If `True` only look for TIGER shapefiles. If `False`, first look
        for CB shapefiles
        (https://www.census.gov/geographies/mapping-files/time-series/geo/carto-boundary-file.html),
        which are more suitable for plotting maps, then fall back on the full
        TIGER files
        (https://www.census.gov/geographies/mapping-files/time-series/geo/tiger-line-file.html)
        only if CB is not available. This is mainly set to `True` only
        when `with_geometry_columns` is also set to `True`. The reason
        is that the additional columns in the shapefiles are different
        in the CB files than in the TIGER files.

    Returns
    -------
        A GeoDataFrame with the original data and an
        added geometry column for each row.
    """
    (
        query_shapefile_scope,
        shapefile_geo_level,
        df_on,
        gdf_on,
    ) = geo_query_from_data_query_inner_geo(year, geo_level)

    # If the query spec has a hard-coded value then we use it.
    if query_shapefile_scope is not None:
        shapefile_scope = query_shapefile_scope

    def individual_shapefile(sub_scope: str, query_year: int) -> gpd.GeoDataFrame:
        """Read the relevant shapefile and add a YEAR column to it."""
        resolution = "5m" if shapefile_geo_level == "nation" else "500k"

        try:
            if tiger_shapefiles_only:
                gdf = __shapefile_reader(query_year).read_shapefile(
                    sub_scope, shapefile_geo_level
                )
            else:
                gdf = __shapefile_reader(query_year).try_cb_tiger_shapefile(
                    sub_scope, shapefile_geo_level, resolution=resolution
                )
            gdf["YEAR"] = query_year
            return gdf
        except cmap.MapException as ex:
            # If there are some years where we can't find a shapefile,
            # skip over it and those rows will not have geometry in the
            # final result.
            logger.info(
                "Unable to load shapefile for scope %s for year %d",
                sub_scope,
                query_year,
                exc_info=ex,
            )
            return gpd.GeoDataFrame()

    # If there is a single defined year then we can load the single
    # shapefile. If not, then we have to load multiple shapefiles,
    # one per year, and concatenate them.
    #
    # Whether there is a single or multiple years, there could also
    # me multiple scopes, e.g. states, for which we have to download
    # shapefiles. If so, by the time we get here, they are encoded in
    # one string with comma separators.
    if isinstance(year, int):
        gdf_shapefile = pd.concat(
            individual_shapefile(sub_scope, year)
            for sub_scope in shapefile_scope.split(",")
        )
        merge_gdf_on = gdf_on
    else:
        gdf_shapefile = pd.concat(
            individual_shapefile(sub_scope, unique_year)
            for unique_year in df_data["YEAR"].unique()
            for sub_scope in shapefile_scope.split(",")
        )

        merge_gdf_on = ["YEAR"] + gdf_on
        df_on = ["YEAR"] + df_on

    if len(gdf_shapefile.index) == 0:
        # None of the years matched, so we add None for geometry to all.
        gdf = gpd.GeoDataFrame(df_data, copy=True)
        gdf.set_geometry([None for _ in gdf.index], inplace=True)
        return gdf

    if "TRACT" in df_data.columns:
        df_data["TRACT"] = df_data["TRACT"].str.ljust(
            6, "0"
        )  # Pre 2010 data has inconsistencies in TRACT string length

    # The nation-level shapefile is missing a column
    # to merge on.
    if shapefile_geo_level == "nation":
        gdf_shapefile["US"] = "1"

    final_data_columns = df_data.columns.tolist()

    # Strip out extra columns unless told to keep them.
    if with_geometry_columns:
        final_data_columns = final_data_columns + [
            col
            for col in gdf_shapefile.columns
            if col not in merge_gdf_on
            and col not in final_data_columns
            and (col != "geometry")
        ]
    else:
        gdf_shapefile = gdf_shapefile[merge_gdf_on + ["geometry"]]

    gdf_data = gdf_shapefile.merge(
        # There is a `NAME` column in the shapefile, so we use suffixes
        # to avoid confusing it with the `NAME` that might be in df_data.
        # Similar for any other conflict in names that are not being merged on.
        df_data,
        how="right",
        left_on=merge_gdf_on,
        right_on=df_on,
        suffixes=("_shapefile", ""),
    )

    # Get the columns we want in a reasonable order matching
    # how they are in the data, with geometry at the end.
    gdf_data = gdf_data[final_data_columns + ["geometry"]]

    # Rearrange columns so geometry is at the end.
    gdf_data = gdf_data[
        [col for col in gdf_data.columns if col != "geometry"] + ["geometry"]
    ]

    return gdf_data


@dataclass
class _ShapefileRoot:
    """A private class to stash the root we will use to cache shapefiles locally."""

    shapefile_root: Optional[Path] = None


__shapefile_root = _ShapefileRoot()
__shapefile_readers: Dict[int, cmap.ShapeReader] = {}


def set_shapefile_path(shapefile_path: Union[Path, None]) -> None:
    """
    Set the path to the directory to cache shapefiles.

    This is where we will cache shapefiles downloaded when
    `with_geometry=True` is passed to :py:func:`~download`.

    Parameters
    ----------
    shapefile_path
        The path to use for caching shapefiles.
    """
    __shapefile_root.shapefile_root = shapefile_path


def get_shapefile_path() -> Union[Path, None]:
    """
    Get the path to the directory to cache shapefiles.

    This is where we will cache shapefiles downloaded when
    `with_geometry=True` is passed to :py:func:`~download`.

    Returns
    -------
        The path to use for caching shapefiles.
    """
    return __shapefile_root.shapefile_root


def __shapefile_reader(year: int):
    reader = __shapefile_readers.get(year, None)

    if reader is None:
        reader = cmap.ShapeReader(
            __shapefile_root.shapefile_root,
            year,
        )

        __shapefile_readers[year] = reader

    return reader


def _identify_counties(gdf_geo: gpd.GeoDataFrame, year: int):
    """
    Take a geodataframe and identify which US counties the supplied geography overlaps.

    Parameters
    ----------
    gdf_geo
        A GeoDataFrame containing polygons within the United States
    year
        The year for which to fetch geometries. We need this
        because they change over time.

    Returns
    -------
        A list of five digit county FIPS codes.
    """
    # Some dataframes will contain the county column already
    if "STATE" in gdf_geo and "COUNTY" in gdf_geo:
        fips_codes = gdf_geo["STATE"] + gdf_geo["COUNTY"]

        return fips_codes.unique().tolist()
    # Otherwise, we load all the US counties and perform an overlap operation
    else:
        reader = __shapefile_readers.get(year)
        us_counties = reader.read_cb_shapefile("us", "county")
        us_counties["FIPS"] = us_counties["STATEFP"] + us_counties["COUNTYFP"]

        county_overlap = us_counties.overlay(gdf_geo, keep_geom_type=False)
        fips_codes = county_overlap["STATEFP"] + county_overlap["COUNTYFP"]

        return fips_codes.unique().tolist()


def clip_water(
    gdf_geo: gpd.GeoDataFrame,
    year: int,
    minimum_area_sq_meters: int = 10000,
    sliver_threshold=0.01,
):
    """
    Remove water from input `GeoDataFrame`.

    Parameters
    ----------
    gdf_geo
        The GeoDataFrame from which we want to remove water
    year
        The year for which to fetch geometries. We need this
        because they change over time.
    minimum_area_sq_meters
        The minimimum size of a water area to be removed

    Returns
    -------
        A GeoDataFrame with the water areas larger than
        the specified threshold removed.
    """
    counties = _identify_counties(gdf_geo, year)
    gdf_water = _retrieve_water(counties, year)

    gdf_without_water = _water_difference(gdf_geo, gdf_water, minimum_area_sq_meters)

    original_crs = gdf_without_water.crs
    gdf_without_water = drop_slivers_from_gdf(
        gdf_without_water.to_crs(epsg=3857), threshold=sliver_threshold
    ).to_crs(original_crs)

    return gdf_without_water


def _retrieve_water(county_fips_codes: list[str], year: int):
    """
    Load `AREAWATER` files from tiger for specified counties.

    Parameters
    ----------
    county_FIPS_codes
        A list of five digit county FIPS codes

    Returns
    -------
        A GeoDataFrame containing the census defined water in the supplied counties
    """
    reader = __shapefile_readers.get(year)

    gdf_water = pd.concat(
        reader.read_shapefile(shapefile_scope=county, geography="areawater")
        for county in county_fips_codes
    )
    # Geo pandas has no concat method, so we convert from a pandas df
    gdf_water = gpd.GeoDataFrame(gdf_water)
    return gdf_water


def _water_difference(
    gdf_geo: gpd.GeoDataFrame, gdf_water: gpd.GeoDataFrame, minimum_area_sq_meters: int
):
    """
    Remove water polygons exceeding minimum size from supplied `GeoDataFrame`.

    Parameters
    ----------
    gdf_geo
        A GeoDataFrame containing polygons within the United States
    gdf_water
        A GeoDataFrame containing census AREAWATER polygons
    minimum_area_sq_meters
        The smallest water polygon to be removed, specified in square meters

    Returns
    -------
        A version of gdf_geo with the water areas removed
    """
    # Combining polygons speeds up the overlay operation
    geo_combined_water = gdf_water[
        gdf_water["AWATER"] >= minimum_area_sq_meters
    ].union_all()
    gdf_combined_water = gpd.GeoDataFrame(geometry=[geo_combined_water])
    gdf_combined_water = gdf_combined_water.set_crs(gdf_water.crs)

    return gdf_geo.overlay(
        gdf_combined_water,
        "difference",
        keep_geom_type=False,
    )



# Copyright (c) 2023 Darren Erik Vengroff
"""Places in Massachusetts."""


ABINGTON_CDP = "00135"
ACUSHNET_CENTER_CDP = "00530"
ADAMS_CDP = "00590"
AGAWAM_TOWN_CITY = "00840"
AMESBURY_TOWN_CITY = "01260"
AMHERST_TOWN_CITY = "01370"
ANDOVER_CDP = "01430"
ARLINGTON_CDP = "01640"
ATHOL_CDP = "02515"
ATTLEBORO_CITY = "02690"
AYER_CDP = "03040"
BALDWINVILLE_CDP = "03355"
BARNSTABLE_TOWN_CITY = "03690"
BARRE_CDP = "03705"
BELCHERTOWN_CDP = "04790"
BELLINGHAM_CDP = "04965"
BELMONT_CDP = "05105"
BEVERLY_CITY = "05595"
BLANDFORD_CDP = "06050"
BLISS_CORNER_CDP = "06170"
BOSTON_CITY = "07000"
BOURNE_CDP = "07140"
BOXFORD_CDP = "07385"
BRAINTREE_TOWN_CITY = "07740"
BREWSTER_CDP = "07945"
BRIDGEWATER_TOWN_CITY = "08130"
BROCKTON_CITY = "09000"
BROOKFIELD_CDP = "09070"
BROOKLINE_CDP = "09210"
BURLINGTON_CDP = "09875"
BUZZARDS_BAY_CDP = "10015"
CAMBRIDGE_CITY = "11000"
CEDAR_CREST_CDP = "11945"
CHATHAM_CDP = "12960"
CHELSEA_CITY = "13205"
CHESHIRE_CDP = "13310"
CHESTER_CDP = "13450"
CHICOPEE_CITY = "13660"
CLINTON_CDP = "14430"
COCHITUATE_CDP = "14570"
CORDAVILLE_CDP = "15340"
DANVERS_CDP = "16285"
DEDHAM_CDP = "16530"
DEERFIELD_CDP = "16635"
DENNIS_CDP = "16740"
DENNIS_PORT_CDP = "16810"
DEVENS_CDP = "16840"
DOVER_CDP = "17370"
DUXBURY_CDP = "17860"
EAST_BROOKFIELD_CDP = "18595"
EAST_DENNIS_CDP = "18840"
EAST_DOUGLAS_CDP = "18875"
EAST_FALMOUTH_CDP = "18980"
EAST_HARWICH_CDP = "19400"
EAST_PEPPERELL_CDP = "20310"
EAST_SANDWICH_CDP = "20380"
EASTHAMPTON_TOWN_CITY = "19370"
EDGARTOWN_CDP = "21115"
ESSEX_CDP = "21815"
EVERETT_CITY = "21990"
FALL_RIVER_CITY = "23000"
FALMOUTH_CDP = "23070"
FISKDALE_CDP = "23840"
FITCHBURG_CITY = "23875"
FORESTDALE_CDP = "24190"
FOXBOROUGH_CDP = "24855"
FRAMINGHAM_CITY = "24960"
FRANKLIN_TOWN_CITY = "25172"
GARDNER_CITY = "25485"
GLOUCESTER_CITY = "26150"
GRANBY_CDP = "26500"
GREAT_BARRINGTON_CDP = "26780"
GREEN_HARBOR_CDP = "27130"
GREENFIELD_CITY = "27060"
GROTON_CDP = "27445"
HANSCOM_AFB_CDP = "28425"
HANSON_CDP = "28460"
HARWICH_CENTER_CDP = "29055"
HARWICH_PORT_CDP = "29090"
HATFIELD_CDP = "29230"
HAVERHILL_CITY = "29405"
HINGHAM_CDP = "30175"
HOLBROOK_CDP = "30420"
HOLLAND_CDP = "30630"
HOLYOKE_CITY = "30840"
HOPEDALE_CDP = "30980"
HOPKINTON_CDP = "31050"
HOUSATONIC_CDP = "31295"
HUDSON_CDP = "31575"
HULL_CDP = "31680"
HUNTINGTON_CDP = "31750"
IPSWICH_CDP = "32275"
KINGSTON_CDP = "33185"
LAWRENCE_CITY = "34550"
LEE_CDP = "34620"
LENOX_CDP = "34935"
LENOX_DALE_CDP = "35040"
LEOMINSTER_CITY = "35075"
LEXINGTON_CDP = "35250"
LITTLETON_COMMON_CDP = "36020"
LONGMEADOW_CDP = "36335"
LOWELL_CITY = "37000"
LUNENBURG_CDP = "37385"
LYNN_CITY = "37490"
LYNNFIELD_CDP = "37595"
MADAKET_CDP = "37735"
MALDEN_CITY = "37875"
MANSFIELD_CENTER_CDP = "38230"
MARBLEHEAD_CDP = "38435"
MARION_CENTER_CDP = "38575"
MARLBOROUGH_CITY = "38715"
MARSHFIELD_CDP = "38820"
MARSHFIELD_HILLS_CDP = "38960"
MASHPEE_NECK_CDP = "39152"
MATTAPOISETT_CENTER_CDP = "39485"
MAYNARD_CDP = "39660"
MEDFIELD_CDP = "39730"
MEDFORD_CITY = "39835"
MELROSE_CITY = "40115"
METHUEN_TOWN_CITY = "40710"
MIDDLEBOROUGH_CENTER_CDP = "40885"
MILFORD_CDP = "41200"
MILLERS_FALLS_CDP = "41410"
MILLIS_CLICQUOT_CDP = "41532"
MILTON_CDP = "41725"
MONOMOSCOY_ISLAND_CDP = "41960"
MONSON_CENTER_CDP = "42180"
MONUMENT_BEACH_CDP = "42670"
NAHANT_CDP = "43615"
NANTUCKET_CDP = "43755"
NEEDHAM_CDP = "44140"
NEW_BEDFORD_CITY = "45000"
NEW_SEABURY_CDP = "45500"
NEWBURYPORT_CITY = "45245"
NEWTON_CITY = "45560"
NORTH_ADAMS_CITY = "46225"
NORTH_ATTLEBOROUGH_TOWN_CITY = "46598"
NORTH_BROOKFIELD_CDP = "47100"
NORTH_EASTHAM_CDP = "47590"
NORTH_FALMOUTH_CDP = "47730"
NORTH_LAKEVILLE_CDP = "48220"
NORTH_PEMBROKE_CDP = "48710"
NORTH_PLYMOUTH_CDP = "48780"
NORTH_SCITUATE_CDP = "49165"
NORTH_SEEKONK_CDP = "49200"
NORTH_WESTPORT_CDP = "49710"
NORTHAMPTON_CITY = "46330"
NORTHBOROUGH_CDP = "46785"
NORTHFIELD_CDP = "47800"
NORTHWEST_HARWICH_CDP = "49700"
NORTON_CENTER_CDP = "50005"
NORWOOD_CDP = "50285"
OAK_BLUFFS_CDP = "50425"
OCEAN_BLUFF_BRANT_ROCK_CDP = "50862"
OCEAN_GROVE_CDP = "50880"
ONSET_CDP = "51160"
ORANGE_CDP = "51230"
ORLEANS_CDP = "51405"
OXFORD_CDP = "51790"
PALMER_TOWN_CITY = "52144"
PEABODY_CITY = "52490"
PEPPERELL_CDP = "52770"
PETERSHAM_CDP = "53085"
PINEHURST_CDP = "53680"
PITTSFIELD_CITY = "53960"
PLYMOUTH_CDP = "54275"
POCASSET_CDP = "54450"
POPPONESSET_CDP = "55070"
POPPONESSET_ISLAND_CDP = "55090"
PROVINCETOWN_CDP = "55535"
QUINCY_CITY = "55745"
RANDOLPH_TOWN_CITY = "56000"
RAYNHAM_CENTER_CDP = "56095"
READING_CDP = "56165"
REVERE_CITY = "56585"
ROCKPORT_CDP = "57845"
ROWLEY_CDP = "58370"
RUSSELL_CDP = "58615"
RUTLAND_CDP = "58790"
SAGAMORE_CDP = "58965"
SALEM_CITY = "59105"
SALISBURY_CDP = "59210"
SANDWICH_CDP = "59700"
SAUGUS_CDP = "60050"
SCITUATE_CDP = "60295"
SEABROOK_CDP = "60505"
SECONSETT_ISLAND_CDP = "60627"
SHARON_CDP = "60820"
SHELBURNE_FALLS_CDP = "61205"
SHIRLEY_CDP = "61555"
SIASCONSET_CDP = "61940"
SMITH_MILLS_CDP = "62300"
SOMERSET_CDP = "62465"
SOMERVILLE_CITY = "62535"
SOUTH_ASHBURNHAM_CDP = "62780"
SOUTH_DEERFIELD_CDP = "63620"
SOUTH_DENNIS_CDP = "63655"
SOUTH_DUXBURY_CDP = "63690"
SOUTH_LANCASTER_CDP = "64460"
SOUTH_YARMOUTH_CDP = "66035"
SOUTHBRIDGE_TOWN_CITY = "63345"
SPENCER_CDP = "66070"
SPRINGFIELD_CITY = "67000"
STONEHAM_CDP = "67700"
STURBRIDGE_CDP = "68120"
SWAMPSCOTT_CDP = "68680"
TAUNTON_CITY = "69170"
TEATICKET_CDP = "69205"
THE_PINEHILLS_CDP = "69585"
TOPSFIELD_CDP = "70115"
TOWNSEND_CDP = "70325"
TURNERS_FALLS_CDP = "70815"
UPTON_CDP = "71445"
VINEYARD_HAVEN_CDP = "71970"
WAKEFIELD_CDP = "72250"
WALPOLE_CDP = "72460"
WALTHAM_CITY = "72600"
WARE_CDP = "72845"
WAREHAM_CENTER_CDP = "73020"
WARREN_CDP = "73055"
WATERTOWN_TOWN_CITY = "73440"
WEBSTER_CDP = "73930"
WELLESLEY_CDP = "74210"
WEST_BROOKFIELD_CDP = "75365"
WEST_CHATHAM_CDP = "75505"
WEST_CONCORD_CDP = "75680"
WEST_DENNIS_CDP = "75820"
WEST_FALMOUTH_CDP = "75960"
WEST_SPRINGFIELD_TOWN_CITY = "77890"
WEST_WAREHAM_CDP = "78410"
WEST_WARREN_CDP = "78445"
WEST_YARMOUTH_CDP = "78795"
WESTBOROUGH_CDP = "75050"
WESTFIELD_CITY = "76030"
WEWEANTIC_CDP = "78850"
WEYMOUTH_TOWN_CITY = "78972"
WHITE_ISLAND_SHORES_CDP = "79390"
WHITINSVILLE_CDP = "79495"
WILBRAHAM_CDP = "79705"
WILLIAMSTOWN_CDP = "79950"
WILMINGTON_CDP = "80195"
WINCHENDON_CDP = "80370"
WINCHESTER_CDP = "80545"
WINTHROP_TOWN_CITY = "81005"
WOBURN_CITY = "81035"
WOODS_HOLE_CDP = "81245"
WORCESTER_CITY = "82000"
YARMOUTH_PORT_CDP = "82595"



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Indiana."""


ADAMS = "001"
ALLEN = "003"
BARTHOLOMEW = "005"
BENTON = "007"
BLACKFORD = "009"
BOONE = "011"
BROWN = "013"
CARROLL = "015"
CASS = "017"
CLARK = "019"
CLAY = "021"
CLINTON = "023"
CRAWFORD = "025"
DAVIESS = "027"
DEKALB = "033"
DEARBORN = "029"
DECATUR = "031"
DELAWARE = "035"
DUBOIS = "037"
ELKHART = "039"
FAYETTE = "041"
FLOYD = "043"
FOUNTAIN = "045"
FRANKLIN = "047"
FULTON = "049"
GIBSON = "051"
GRANT = "053"
GREENE = "055"
HAMILTON = "057"
HANCOCK = "059"
HARRISON = "061"
HENDRICKS = "063"
HENRY = "065"
HOWARD = "067"
HUNTINGTON = "069"
JACKSON = "071"
JASPER = "073"
JAY = "075"
JEFFERSON = "077"
JENNINGS = "079"
JOHNSON = "081"
KNOX = "083"
KOSCIUSKO = "085"
LAGRANGE = "087"
LAPORTE = "091"
LAKE = "089"
LAWRENCE = "093"
MADISON = "095"
MARION = "097"
MARSHALL = "099"
MARTIN = "101"
MIAMI = "103"
MONROE = "105"
MONTGOMERY = "107"
MORGAN = "109"
NEWTON = "111"
NOBLE = "113"
OHIO = "115"
ORANGE = "117"
OWEN = "119"
PARKE = "121"
PERRY = "123"
PIKE = "125"
PORTER = "127"
POSEY = "129"
PULASKI = "131"
PUTNAM = "133"
RANDOLPH = "135"
RIPLEY = "137"
RUSH = "139"
SCOTT = "143"
SHELBY = "145"
SPENCER = "147"
ST_JOSEPH = "141"
STARKE = "149"
STEUBEN = "151"
SULLIVAN = "153"
SWITZERLAND = "155"
TIPPECANOE = "157"
TIPTON = "159"
UNION = "161"
VANDERBURGH = "163"
VERMILLION = "165"
VIGO = "167"
WABASH = "169"
WARREN = "171"
WARRICK = "173"
WASHINGTON = "175"
WAYNE = "177"
WELLS = "179"
WHITE = "181"
WHITLEY = "183"



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Missouri."""


ADAIR = "001"
ANDREW = "003"
ATCHISON = "005"
AUDRAIN = "007"
BARRY = "009"
BARTON = "011"
BATES = "013"
BENTON = "015"
BOLLINGER = "017"
BOONE = "019"
BUCHANAN = "021"
BUTLER = "023"
CALDWELL = "025"
CALLAWAY = "027"
CAMDEN = "029"
CAPE_GIRARDEAU = "031"
CARROLL = "033"
CARTER = "035"
CASS = "037"
CEDAR = "039"
CHARITON = "041"
CHRISTIAN = "043"
CLARK = "045"
CLAY = "047"
CLINTON = "049"
COLE = "051"
COOPER = "053"
CRAWFORD = "055"
DADE = "057"
DALLAS = "059"
DAVIESS = "061"
DEKALB = "063"
DENT = "065"
DOUGLAS = "067"
DUNKLIN = "069"
FRANKLIN = "071"
GASCONADE = "073"
GENTRY = "075"
GREENE = "077"
GRUNDY = "079"
HARRISON = "081"
HENRY = "083"
HICKORY = "085"
HOLT = "087"
HOWARD = "089"
HOWELL = "091"
IRON = "093"
JACKSON = "095"
JASPER = "097"
JEFFERSON = "099"
JOHNSON = "101"
KNOX = "103"
LACLEDE = "105"
LAFAYETTE = "107"
LAWRENCE = "109"
LEWIS = "111"
LINCOLN = "113"
LINN = "115"
LIVINGSTON = "117"
MACON = "121"
MADISON = "123"
MARIES = "125"
MARION = "127"
MCDONALD = "119"
MERCER = "129"
MILLER = "131"
MISSISSIPPI = "133"
MONITEAU = "135"
MONROE = "137"
MONTGOMERY = "139"
MORGAN = "141"
NEW_MADRID = "143"
NEWTON = "145"
NODAWAY = "147"
OREGON = "149"
OSAGE = "151"
OZARK = "153"
PEMISCOT = "155"
PERRY = "157"
PETTIS = "159"
PHELPS = "161"
PIKE = "163"
PLATTE = "165"
POLK = "167"
PULASKI = "169"
PUTNAM = "171"
RALLS = "173"
RANDOLPH = "175"
RAY = "177"
REYNOLDS = "179"
RIPLEY = "181"
SALINE = "195"
SCHUYLER = "197"
SCOTLAND = "199"
SCOTT = "201"
SHANNON = "203"
SHELBY = "205"
ST_CHARLES = "183"
ST_CLAIR = "185"
ST_FRANCOIS = "187"
ST_LOUIS = "189"
ST_LOUIS_CITY = "510"
STE_GENEVIEVE = "186"
STODDARD = "207"
STONE = "209"
SULLIVAN = "211"
TANEY = "213"
TEXAS = "215"
VERNON = "217"
WARREN = "219"
WASHINGTON = "221"
WAYNE = "223"
WEBSTER = "225"
WORTH = "227"
WRIGHT = "229"



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Utah."""


BEAVER = "001"
BOX_ELDER = "003"
CACHE = "005"
CARBON = "007"
DAGGETT = "009"
DAVIS = "011"
DUCHESNE = "013"
EMERY = "015"
GARFIELD = "017"
GRAND = "019"
IRON = "021"
JUAB = "023"
KANE = "025"
MILLARD = "027"
MORGAN = "029"
PIUTE = "031"
RICH = "033"
SALT_LAKE = "035"
SAN_JUAN = "037"
SANPETE = "039"
SEVIER = "041"
SUMMIT = "043"
TOOELE = "045"
UINTAH = "047"
UTAH = "049"
WASATCH = "051"
WASHINGTON = "053"
WAYNE = "055"
WEBER = "057"



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Washington."""


ADAMS = "001"
ASOTIN = "003"
BENTON = "005"
CHELAN = "007"
CLALLAM = "009"
CLARK = "011"
COLUMBIA = "013"
COWLITZ = "015"
DOUGLAS = "017"
FERRY = "019"
FRANKLIN = "021"
GARFIELD = "023"
GRANT = "025"
GRAYS_HARBOR = "027"
ISLAND = "029"
JEFFERSON = "031"
KING = "033"
KITSAP = "035"
KITTITAS = "037"
KLICKITAT = "039"
LEWIS = "041"
LINCOLN = "043"
MASON = "045"
OKANOGAN = "047"
PACIFIC = "049"
PEND_OREILLE = "051"
PIERCE = "053"
SAN_JUAN = "055"
SKAGIT = "057"
SKAMANIA = "059"
SNOHOMISH = "061"
SPOKANE = "063"
STEVENS = "065"
THURSTON = "067"
WAHKIAKUM = "069"
WALLA_WALLA = "071"
WHATCOM = "073"
WHITMAN = "075"
YAKIMA = "077"



# Copyright (c) 2022 Darren Erik Vengroff
"""Utilities for loading census data."""
from logging import getLogger
from typing import Any, Mapping, Optional, Union, Tuple

import pandas as pd
import requests

from censusdis.impl.exceptions import CensusApiException

logger = getLogger(__name__)


class _CertificateManager:
    """Manage the certificates and verification flags used when we make calls to the U.S. Census servers."""

    _have_a_singletop_certificate_manager = False

    def __init__(
        self,
        *,
        data_verify: Union[bool, str] = True,
        data_cert: Optional[Union[str, Tuple[str, str]]] = None,
        map_verify: Union[bool, str] = True,
        map_cert: Optional[Union[str, Tuple[str, str]]] = None,
    ):
        """
        Manage the certificates and verification flags used when we make calls to the U.S. Census servers.

        Parameters
        ----------
        data_verify
            Value to pass to `requests.get` in the `verify=` argument for data API calls to `https://api.census.gov`.
        data_cert
            Value to pass to `requests.get` in the `cert=` argument for data API calls to `https://api.census.gov`.
        map_verify
            Value to pass to `requests.get` in the `verify=` argument for getting map data with calls to
            `https://www2.census.gov`.
        map_cert
            Value to pass to `requests.get` in the `cert=` argument for getting map data with calls to
            `https://www2.census.gov`.
        """
        if self._have_a_singletop_certificate_manager:
            raise ValueError(
                "Cannot create a CertificateManager. Please use `censusdis.data.certificates`."
            )

        self._data_verify = data_verify
        self._data_cert = data_cert
        self._map_verify = map_verify
        self._map_cert = map_cert

        self._have_a_singletop_certificate_manager = True

    @property
    def data_verify(self) -> Union[bool, str]:
        """Value to pass to `requests.get` in the `verify=` argument for data from `https://api.census.gov`."""
        return self._data_verify

    @property
    def data_cert(self) -> Union[str, Tuple[str, str], None]:
        """Value to pass to `requests.get` in the `cert=` argument for data from `https://api.census.gov`."""
        return self._data_cert

    @property
    def map_verify(self) -> Union[bool, str]:
        """Value to pass to `requests.get` in the `verify=` argument for maps from `https://www2.census.gov`."""
        return self._map_verify

    @property
    def map_cert(self) -> Union[str, Tuple[str, str], None]:
        """Value to pass to `requests.get` in the `cert=` argument for maps from `https://www2.census.gov`."""
        return self._map_cert

    def use(
        self,
        *,
        data_verify: Union[bool, str] = True,
        data_cert: Optional[Union[str, Tuple[str, str]]] = None,
        map_verify: Union[bool, str] = True,
        map_cert: Optional[Union[str, Tuple[str, str]]] = None,
    ) -> "_CertificateManagerContext":
        """
        Set certificates and verification flags globally or within a context.

        If you want to set up certificate handling globally, you can just call this
        method alone, for example:

            import censusdis.data as ced

            ced.certificates.use(data_verify=False, map_verify=False)

        will turn off certificate verification for all data and map calls. This can by useful
        in a notebook environment, where you want to set up how certificates are handled once
        at the top of the notebook.

        If you want the effects to only be temporary, you can use a context manager with a `with`
        statement as follows::

            import censusdis.data as ced

            with ced.certificates.use(data_verify=False, map_verify=False):
                # No verification will be performed here.
                df = ced.download(...)

            # Upon exiting the context, verification is back on.
            df = ced.download(...)

        Parameters
        ----------
        data_verify
            Value to pass to `requests.get` in the `verify=` argument for data API calls to `https://api.census.gov`.
        data_cert
            Value to pass to `requests.get` in the `cert=` argument for data API calls to `https://api.census.gov`.
        map_verify
            Value to pass to `requests.get` in the `verify=` argument for getting map data with calls to
            `https://www2.census.gov`.
        map_cert
            Value to pass to `requests.get` in the `cert=` argument for getting map data with calls to
            `https://www2.census.gov`.
        """
        context = _CertificateManagerContext(self)
        self._data_verify = data_verify
        self._data_cert = data_cert
        self._map_verify = map_verify
        self._map_cert = map_cert
        return context


class _CertificateManagerContext:
    def __init__(self, certificate_manager: _CertificateManager):
        self._certificate_manager = certificate_manager
        self._data_verify = certificate_manager.data_verify
        self._data_cert = certificate_manager.data_cert
        self._map_verify = certificate_manager.map_verify
        self._map_cert = certificate_manager.map_cert

    def __enter__(self):
        pass

    def __exit__(self, type_, value, traceback):
        self._certificate_manager._data_verify = self._data_verify
        self._certificate_manager._data_cert = self._data_cert
        self._certificate_manager._map_verify = self._map_verify
        self._certificate_manager._map_cert = self._map_cert


certificates = _CertificateManager()
"""
A container for the certificates and verification flags used when we make calls to the U.S. Census servers.

Unless you are working behind a security proxy or firewall that manipulates certificates in
some way, you will never have to use this.

If you would not normally use the `verify=` or `cert=` arguments when using `requests.get` then
you need not worry about this. If you would, then use the values you would pass for accessing
`https://api.census.gov` or `https://www2.census.gov`.
"""


def json_from_url(url: str, params: Optional[Mapping[str, str]] = None) -> Any:
    """Get json from a URL."""
    request = requests.get(
        url, params=params, cert=certificates.data_cert, verify=certificates.data_verify
    )

    if request.status_code == 200:
        try:
            parsed_json = request.json()
            return parsed_json
        except requests.exceptions.JSONDecodeError:
            logger.debug(f"API call got 200 with unparseable JSON:\n{request.text}")
            if (
                "You included a key with this request, however, it is not valid."
                in request.text
            ):
                message = f"Census API request to {request.url} failed because your key is invalid."
            else:
                message = f"Census API request to {request.url} failed. Unable to parse returned JSON:\n{request.text}"
            raise CensusApiException(message)

    # Do our best to tell the user something informative.
    message = f"Census API request to {request.url} failed with status {request.status_code}. {request.text}"
    logger.debug(message)
    raise CensusApiException(message)


def data_from_url(url: str, params: Optional[Mapping[str, str]] = None) -> pd.DataFrame:
    """Get json from a URL and parse into a data frame."""
    logger.info(f"Downloading data from {url} with {params}.")

    parsed_json = json_from_url(url, params)

    return _df_from_census_json(parsed_json)


def _df_from_census_json(parsed_json):
    if (
        isinstance(parsed_json, list)
        and len(parsed_json) >= 1
        and isinstance(parsed_json[0], list)
    ):
        return pd.DataFrame(
            parsed_json[1:],
            columns=[
                c.upper()
                .replace(" ", "_")
                .replace("-", "_")
                .replace("/", "_")
                .replace("(", "")
                .replace(")", "")
                for c in parsed_json[0]
            ],
        )

    raise CensusApiException(
        f"Expected json data to be a list of lists, not a {type(parsed_json)}"
    )



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Nebraska."""


ADAMS = "001"
ANTELOPE = "003"
ARTHUR = "005"
BANNER = "007"
BLAINE = "009"
BOONE = "011"
BOX_BUTTE = "013"
BOYD = "015"
BROWN = "017"
BUFFALO = "019"
BURT = "021"
BUTLER = "023"
CASS = "025"
CEDAR = "027"
CHASE = "029"
CHERRY = "031"
CHEYENNE = "033"
CLAY = "035"
COLFAX = "037"
CUMING = "039"
CUSTER = "041"
DAKOTA = "043"
DAWES = "045"
DAWSON = "047"
DEUEL = "049"
DIXON = "051"
DODGE = "053"
DOUGLAS = "055"
DUNDY = "057"
FILLMORE = "059"
FRANKLIN = "061"
FRONTIER = "063"
FURNAS = "065"
GAGE = "067"
GARDEN = "069"
GARFIELD = "071"
GOSPER = "073"
GRANT = "075"
GREELEY = "077"
HALL = "079"
HAMILTON = "081"
HARLAN = "083"
HAYES = "085"
HITCHCOCK = "087"
HOLT = "089"
HOOKER = "091"
HOWARD = "093"
JEFFERSON = "095"
JOHNSON = "097"
KEARNEY = "099"
KEITH = "101"
KEYA_PAHA = "103"
KIMBALL = "105"
KNOX = "107"
LANCASTER = "109"
LINCOLN = "111"
LOGAN = "113"
LOUP = "115"
MADISON = "119"
MCPHERSON = "117"
MERRICK = "121"
MORRILL = "123"
NANCE = "125"
NEMAHA = "127"
NUCKOLLS = "129"
OTOE = "131"
PAWNEE = "133"
PERKINS = "135"
PHELPS = "137"
PIERCE = "139"
PLATTE = "141"
POLK = "143"
RED_WILLOW = "145"
RICHARDSON = "147"
ROCK = "149"
SALINE = "151"
SARPY = "153"
SAUNDERS = "155"
SCOTTS_BLUFF = "157"
SEWARD = "159"
SHERIDAN = "161"
SHERMAN = "163"
SIOUX = "165"
STANTON = "167"
THAYER = "169"
THOMAS = "171"
THURSTON = "173"
VALLEY = "175"
WASHINGTON = "177"
WAYNE = "179"
WEBSTER = "181"
WHEELER = "183"
YORK = "185"



# Copyright (c) 2023 Darren Erik Vengroff
"""Places in Maine."""


ALFRED_CDP = "00695"
ANSON_CDP = "01360"
ASHLAND_CDP = "01675"
AUBURN_CITY = "02060"
AUGUSTA_CITY = "02100"
BANGOR_CITY = "02795"
BAR_HARBOR_CDP = "02830"
BATH_CITY = "03355"
BELFAST_CITY = "03950"
BERWICK_CDP = "04685"
BETHEL_CDP = "04790"
BIDDEFORD_CITY = "04860"
BINGHAM_CDP = "04965"
BLAINE_CDP = "05350"
BLUE_HILL_CDP = "05665"
BOOTHBAY_HARBOR_CDP = "06085"
BOWDOINHAM_CDP = "06330"
BRADLEY_CDP = "06645"
BREWER_CITY = "06925"
BRIDGTON_CDP = "07135"
BROWNVILLE_JUNCTION_CDP = "08360"
BRUNSWICK_CDP = "08395"
BUCKSPORT_CDP = "08780"
CALAIS_CITY = "09585"
CAMDEN_CDP = "09690"
CAPE_NEDDICK_CDP = "10320"
CARIBOU_CITY = "10565"
CASCO_CDP = "11090"
CASTINE_CDP = "11230"
CHISHOLM_CDP = "12770"
CLINTON_CDP = "13435"
CORINNA_CDP = "14275"
CORNISH_CDP = "14450"
COUSINS_ISLAND_CDP = "14800"
CUMBERLAND_CENTER_CDP = "15500"
DAMARISCOTTA_CDP = "16200"
DANFORTH_CDP = "16375"
DEXTER_CDP = "17495"
DIXFIELD_CDP = "17705"
DOVER_FOXCROFT_CDP = "18230"
DUNSTAN_CDP = "18965"
EAGLE_LAKE_CDP = "19385"
EAST_MILLINOCKET_CDP = "21065"
EASTPORT_CITY = "21730"
ELLSWORTH_CITY = "23200"
FAIRFIELD_CDP = "24285"
FALMOUTH_CDP = "24480"
FALMOUTH_FORESIDE_CDP = "24530"
FARMINGDALE_CDP = "24635"
FARMINGTON_CDP = "24740"
FORT_FAIRFIELD_CDP = "25580"
FORT_KENT_CDP = "25720"
FREEPORT_CDP = "26490"
FRYEBURG_CDP = "26875"
GARDINER_CITY = "27085"
GORHAM_CDP = "28205"
GRAND_ISLE_CDP = "28555"
GRAY_CDP = "28835"
GREENE_CDP = "29220"
GREENVILLE_CDP = "29500"
GUILFORD_CDP = "30060"
HALLOWELL_CITY = "30550"
HAMPDEN_CDP = "30760"
HARTLAND_CDP = "31705"
HOULTON_CDP = "33945"
HOWLAND_CDP = "34225"
ISLAND_FALLS_CDP = "35030"
JONESPORT_CDP = "35975"
KENNEBUNK_CDP = "36500"
KENNEBUNKPORT_CDP = "36710"
KEZAR_FALLS_CDP = "36885"
KINGFIELD_CDP = "36990"
KITTERY_CDP = "37235"
KITTERY_POINT_CDP = "37375"
LAKE_ARROWHEAD_CDP = "37795"
LEWISTON_CITY = "38740"
LIMESTONE_CDP = "39265"
LINCOLN_CDP = "39440"
LISBON_CDP = "40000"
LISBON_FALLS_CDP = "40105"
LITTLE_FALLS_CDP = "40350"
LITTLEJOHN_ISLAND_CDP = "40470"
LIVERMORE_FALLS_CDP = "40735"
LUBEC_CDP = "41575"
MACHIAS_CDP = "41925"
MADAWASKA_CDP = "42485"
MADISON_CDP = "42625"
MAPLETON_CDP = "43220"
MARS_HILL_CDP = "43675"
MATTAWAMKEAG_CDP = "44235"
MECHANIC_FALLS_CDP = "44620"
MEXICO_CDP = "45250"
MILBRIDGE_CDP = "45565"
MILFORD_CDP = "45635"
MILLINOCKET_CDP = "45845"
MILO_CDP = "45985"
NAPLES_CDP = "48050"
NEWCASTLE_CDP = "48610"
NEWPORT_CDP = "49100"
NORRIDGEWOCK_CDP = "49800"
NORTH_ANSON_CDP = "49975"
NORTH_BERWICK_CDP = "50290"
NORTH_WINDHAM_CDP = "53685"
NORTHEAST_HARBOR_CDP = "51095"
NORWAY_CDP = "53965"
OAK_HILL_CDP = "54490"
OAKFIELD_CDP = "54350"
OAKLAND_CDP = "54525"
OLD_ORCHARD_BEACH_CDP = "55120"
OLD_TOWN_CITY = "55225"
ORONO_CDP = "55575"
OXFORD_CDP = "56240"
PATTEN_CDP = "57115"
PITTSFIELD_CDP = "58970"
PORTLAND_CITY = "60545"
PRESQUE_ISLE_CITY = "60825"
RANDOLPH_CDP = "61735"
RANGELEY_CDP = "61805"
RICHMOND_CDP = "62610"
ROCKLAND_CITY = "63590"
RUMFORD_CDP = "64255"
SABATTUS_CDP = "64535"
SACO_CITY = "64675"
SANFORD_CITY = "65725"
SEARSPORT_CDP = "66600"
SKOWHEGAN_CDP = "68875"
SOUTH_BERWICK_CDP = "69995"
SOUTH_ELIOT_CDP = "70660"
SOUTH_PARIS_CDP = "71780"
SOUTH_PORTLAND_CITY = "71990"
SOUTH_WINDHAM_CDP = "72900"
SOUTHWEST_HARBOR_CDP = "72830"
STANDISH_CDP = "73635"
STEEP_FALLS_CDP = "73985"
THOMASTON_CDP = "76330"
TOPSHAM_CDP = "76925"
TURNER_CDP = "77765"
UNITY_CDP = "78220"
VAN_BUREN_CDP = "78535"
VANCEBORO_CDP = "78640"
VEAZIE_CDP = "78815"
WALDOBORO_CDP = "79515"
WASHBURN_CDP = "80250"
WATERVILLE_CITY = "80740"
WEST_KENNEBUNK_CDP = "83330"
WESTBROOK_CITY = "82105"
WILTON_CDP = "85815"
WINSLOW_CDP = "86480"
WINTER_HARBOR_CDP = "86620"
WINTERPORT_CDP = "86725"
WINTHROP_CDP = "86935"
WISCASSET_CDP = "87040"
WOODLAND_CDP = "87250"
YARMOUTH_CDP = "87810"
YORK_HARBOR_CDP = "88160"



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Minnesota."""


AITKIN = "001"
ANOKA = "003"
BECKER = "005"
BELTRAMI = "007"
BENTON = "009"
BIG_STONE = "011"
BLUE_EARTH = "013"
BROWN = "015"
CARLTON = "017"
CARVER = "019"
CASS = "021"
CHIPPEWA = "023"
CHISAGO = "025"
CLAY = "027"
CLEARWATER = "029"
COOK = "031"
COTTONWOOD = "033"
CROW_WING = "035"
DAKOTA = "037"
DODGE = "039"
DOUGLAS = "041"
FARIBAULT = "043"
FILLMORE = "045"
FREEBORN = "047"
GOODHUE = "049"
GRANT = "051"
HENNEPIN = "053"
HOUSTON = "055"
HUBBARD = "057"
ISANTI = "059"
ITASCA = "061"
JACKSON = "063"
KANABEC = "065"
KANDIYOHI = "067"
KITTSON = "069"
KOOCHICHING = "071"
LAC_QUI_PARLE = "073"
LAKE = "075"
LAKE_OF_THE_WOODS = "077"
LE_SUEUR = "079"
LINCOLN = "081"
LYON = "083"
MAHNOMEN = "087"
MARSHALL = "089"
MARTIN = "091"
MCLEOD = "085"
MEEKER = "093"
MILLE_LACS = "095"
MORRISON = "097"
MOWER = "099"
MURRAY = "101"
NICOLLET = "103"
NOBLES = "105"
NORMAN = "107"
OLMSTED = "109"
OTTER_TAIL = "111"
PENNINGTON = "113"
PINE = "115"
PIPESTONE = "117"
POLK = "119"
POPE = "121"
RAMSEY = "123"
RED_LAKE = "125"
REDWOOD = "127"
RENVILLE = "129"
RICE = "131"
ROCK = "133"
ROSEAU = "135"
SCOTT = "139"
SHERBURNE = "141"
SIBLEY = "143"
ST_LOUIS = "137"
STEARNS = "145"
STEELE = "147"
STEVENS = "149"
SWIFT = "151"
TODD = "153"
TRAVERSE = "155"
WABASHA = "157"
WADENA = "159"
WASECA = "161"
WASHINGTON = "163"
WATONWAN = "165"
WILKIN = "167"
WINONA = "169"
WRIGHT = "171"
YELLOW_MEDICINE = "173"



# Copyright (c) 2023 Darren Erik Vengroff
"""Counties in Vermont."""


ADDISON = "001"
BENNINGTON = "003"
CALEDONIA = "005"
CHITTENDEN = "007"
ESSEX = "009"
FRANKLIN = "011"
GRAND_ISLE = "013"
LAMOILLE = "015"
ORANGE = "017"
ORLEANS = "019"
RUTLAND = "021"
WASHINGTON = "023"
WINDHAM = "025"
WINDSOR = "027"
